import atexit
from dataclasses import dataclass, fields
import sys
import os.path
from datetime import datetime, timedelta
import contextlib
import asyncio
from typing import OrderedDict
from omegaconf import OmegaConf
import psutil
import rich.progress
import rich.logging
from rich.table import Table
from rich.text import Text

progress_bar = progress_task = None

_progress_task_names = []
_progress_task_names_orig = []

_start_time = datetime.now()
_prev_disk_io = None, None


def init_progress_bar():
    global progress_console, progress_bar, progress_task
    progress_console = rich.console.Console(file=sys.stdout, highlight=False)
    progress_bar = rich.progress.Progress(
        rich.progress.SpinnerColumn(),
        "[yellow]{task.fields[elapsed_time]}[/yellow]",
        "[bold]{task.description}[/bold]",
        rich.progress.SpinnerColumn(),
        "{task.fields[command]}",
        rich.progress.TimeElapsedColumn(),
        "{task.fields[cpu_info]}",
        refresh_per_second=2,
        console=progress_console,
        transient=True)

    progress_task = progress_bar.add_task("stimela", command="starting", cpu_info=" ", elapsed_time="", start=True)
    progress_bar.__enter__()
    atexit.register(destroy_progress_bar)
    return progress_bar, progress_console

def destroy_progress_bar():
    global progress_bar
    if progress_bar is not None:
        progress_bar.__exit__(None, None, None)
        progress_bar = None

@contextlib.contextmanager
def declare_subtask(subtask_name):
    _progress_task_names.append(subtask_name)
    _progress_task_names_orig.append(subtask_name)
    update_process_status(description='.'.join(_progress_task_names))
    try:
        yield subtask_name
    finally:
        _progress_task_names.pop(-1)
        _progress_task_names_orig.pop(-1)
        update_process_status(progress_task,
                              description='.'.join(_progress_task_names))


def declare_subtask_attributes(*args, **kw):
    attrs = [str(x) for x in args] + \
        [f"{key} {value}" for key, value in kw.items()]
    attrs = ', '.join(attrs)
    _progress_task_names[-1] = f"{_progress_task_names_orig[-1]}\[{attrs}]"
    update_process_status(description='.'.join(_progress_task_names))


class _CommandContext(object):
    def __init__(self, command):
        self.command = command
        update_process_status(command=command)
    def ctrl_c(self):
        update_process_status(command=f"{self.command}(^C)")


@contextlib.contextmanager
def declare_subcommand(command):
    update_process_status(command=command)
    progress_bar and progress_bar.reset(progress_task)
    try:
        yield _CommandContext(command)
    finally:
        update_process_status(command="")


@dataclass
class TaskStatsDatum(object):
    cpu: float          = 0
    mem_used: float     = 0
    mem_total: float    = 0
    load: float         = 0
    read_count: int     = 0
    read_gb: float      = 0
    read_gbps: float    = 0
    read_ms: float      = 0
    write_count: int    = 0
    write_gb: float     = 0 
    write_gbps: float   = 0 
    write_ms: float     = 0
    num_samples: int    = 0

    def add(self, other: "TaskStatsDatum"):
        for f in _taskstats_sample_names:
            setattr(self, f, getattr(self, f) + getattr(other, f))

    def peak(self, other: "TaskStatsDatum"):
        for f in _taskstats_sample_names:
            setattr(self, f, max(getattr(self, f), getattr(other, f)))

    def averaged(self):
        avg = TaskStatsDatum(num_samples=1)
        for f in _taskstats_sample_names:
            setattr(avg, f, getattr(self, f) / self.num_samples)
        return avg


_taskstats_sample_names = [f.name for f in fields(TaskStatsDatum)]

_taskstats = OrderedDict()
_task_start_time = OrderedDict()


def collect_stats():
    """Returns dictionary of per-task stats (elapsed time, sums, peaks)"""
    return _taskstats


def add_missing_stats(stats):
    """Adds stats that wren't recorded into dictionary"""
    for key, value in stats.items():
        if key not in _taskstats:
            _taskstats[key] = value


def stats_field_names():
    return _taskstats_sample_names


def update_stats(now: datetime, sample: TaskStatsDatum):
    key1, key2 = tuple(_progress_task_names_orig), tuple(_progress_task_names)

    _, sum, peak = _taskstats.setdefault(key1, [0, TaskStatsDatum(), TaskStatsDatum()])
    sum.add(sample)
    peak.peak(sample)
    start = _task_start_time.setdefault(key1, now)
    _taskstats[key1][0] = (now - start).total_seconds()

    if key2 != key1:
        _, sum, peak = _taskstats.setdefault(key2, [0, TaskStatsDatum(), TaskStatsDatum()])
        sum.add(sample)
        peak.peak(sample)
        start = _task_start_time.setdefault(key2, now)
        _taskstats[key2][0] = (now - start).total_seconds()


def update_process_status(command=None, description=None):
    # elapsed time since start
    now = datetime.now()
    elapsed = str(now - _start_time).split('.', 1)[0]

    # form up sample datum
    s = TaskStatsDatum(num_samples=1)
    # CPU and memory
    s.cpu = psutil.cpu_percent()
    mem = psutil.virtual_memory()
    s.mem_used = round(mem.total*mem.percent/100 / 2**30)
    s.mem_total = round(mem.total / 2**30)
    # load
    s.load, _, _ = psutil.getloadavg()

    # get disk I/O stats
    disk_io = psutil.disk_io_counters()
    global _prev_disk_io
    prev_io, prev_time = _prev_disk_io 
    if prev_io is not None:
        delta = (now - prev_time).total_seconds()
        io = {}
        for key in 'read_bytes', 'read_count', 'read_time', 'write_bytes', 'write_count', 'write_time':
            io[key] = getattr(disk_io, key) - getattr(prev_io, key)
        s.read_count = io['read_count']
        s.write_count = io['write_count']
        s.read_gb = io['read_bytes']/2**30
        s.write_gb = io['write_bytes']/2**30
        s.read_gbps = s.read_gb / delta
        s.write_gbps = s.write_gb / delta
        s.read_ms = io['read_time'] 
        s.write_ms = io['write_time']
    else:
        io = None
    _prev_disk_io = disk_io, now

    # if a progress bar exists, update it
    if progress_bar is not None:
        if io is not None:
            ioinfo = f"|R [green]{s.read_count:-4}[/green] [green]{s.read_gbps:2.2f}[/green]G [green]{s.read_ms:4}[/green]ms" + \
                    f"|W [green]{s.write_count:-4}[/green] [green]{s.write_gbps:2.2f}[/green]G [green]{s.write_ms:4}[/green]ms "
        else:
            ioinfo = ""

        updates = dict(elapsed_time=elapsed,
                    cpu_info=f"CPU [green]{s.cpu:2.1f}%[/green]|RAM [green]{round(s.mem_used):3}[/green]/[green]{round(s.mem_total)}[/green]G|Load [green]{s.load:2.1f}[/green]{ioinfo}")
        if command is not None:
            updates['command'] = command
        if description is not None:
            updates['description'] = description
        progress_bar.update(progress_task, **updates)

    # update stats
    update_stats(now, s)


async def run_process_status_update():
    if progress_bar:
        with contextlib.suppress(asyncio.CancelledError):
            while True:
                update_process_status()
                await asyncio.sleep(1)

_printed_stats = dict(
    cpu="CPU %",
    mem_used="Mem GB",
    load="Load",
    read_gbps="R GB/s",
    write_gbps="W GB/s")

# these stats are written as sums
_sum_stats = ("read_count", "read_gb", "read_ms", "write_count", "write_gb", "write_ms")

def render_profiling_summary(stats, max_depth, unroll_loops=False):

    table_avg = Table(title=Text("\naverages & total I/O", style="bold"))
    table_avg.add_column("")
    table_avg.add_column("time hms", justify="right")

    table_peak = Table(title=Text("\npeaks", style="bold"))
    table_peak.add_column("")
    table_peak.add_column("time hms", justify="right")
    
    for f, label in _printed_stats.items():
        table_avg.add_column(label, justify="right")
        table_peak.add_column(label, justify="right")

    table_avg.add_column("R GB", justify="right")
    table_avg.add_column("W GB", justify="right")

    for name, (elapsed, sum, peak) in stats.items():
        if name and len(name) <= max_depth:
            # skip loop iterations, if not unrolling loops 
            if not unroll_loops and any(n.endswith("]") for n in name):
                continue
            secs, mins, hours = elapsed % 60, int(elapsed // 60) % 60, int(elapsed // 3600)
            tstr = f"{hours:d}:{mins:02d}:{secs:04.1f}"
            avg = sum.averaged()
            avg_row = [".".join(name), tstr]
            peak_row = avg_row.copy()
            for f, label in _printed_stats.items():
                if f != "num_samples":
                    avg_row.append(f"{getattr(avg, f):.2f}")
                    peak_row.append(f"{getattr(peak, f):.2f}")
            avg_row += [f"{sum.read_gb:.2f}", f"{sum.write_gb:.2f}"]
            table_avg.add_row(*avg_row)
            table_peak.add_row(*peak_row)

    progress_console.rule("profiling results")
    destroy_progress_bar()
    from rich.columns import Columns
    # progress_console.print(table_avg, justify="center") 
    # progress_console.print(table_peak, justify="center")  
    # progress_console.print(Columns((table_avg, table_peak)), justify="center") 

    with progress_console.capture() as capture:
        progress_console.print(Columns((table_avg, table_peak)), justify="center") 

    text = capture.get()

    return text

    
# from rich.console import Console
# console = Console()
# with console.capture() as capture:
#     console.print("[bold red]Hello[/] World")
# str_output = capture.get()

def save_profiling_stats(log, print_depth=2, unroll_loops=False):
    from . import stimelogging
    
    stats = collect_stats()
    summary = render_profiling_summary(stats, print_depth, unroll_loops=unroll_loops)
    if print_depth:
        print(summary)

    filename = os.path.join(stimelogging.get_logfile_dir(log) or '.', "stimela.stats.full")

    stats_dict = OmegaConf.create()

    for name, (elapsed, sum, peak) in stats.items():
        if name:
            name = '.'.join(name)
            avg = sum.averaged()
            davg = {f: getattr(avg, f) for f in _taskstats_sample_names}
            dpeak = {f: getattr(peak, f) for f in _taskstats_sample_names}
            dsum = {f: getattr(sum, f) for f in _sum_stats}

            stats_dict[name] = dict(elapsed=elapsed, avg=davg, peak=dpeak, total=dsum)

    OmegaConf.save(stats_dict, filename)

    log.info(f"saved full profiling stats to {filename}")

    filename = os.path.join(stimelogging.get_logfile_dir(log) or '.', "stimela.stats.summary.txt")
    open(filename, "wt").write(summary)

    log.info(f"saved summary to {filename}")



import logging, datetime, resource

from typing import Dict, Optional, Any

import stimela
from stimela.kitchen.cab import Cab
from stimela.utils.xrun_asyncio import xrun
from stimela.schedulers.slurm import SlurmBatch
from stimela.schedulers import SlurmBatch
from stimela.exceptions import StimelaProcessRuntimeError


def update_rlimits(rlimits: Dict[str, Any], log: logging.Logger):
    for name, limit in rlimits.items():
        rname = f"RLIMIT_{name}"
        if not hasattr(resource, rname):
            raise StimelaProcessRuntimeError(f"unknown resource limit 'opts.rlimits.{name}'")
        rconst = getattr(resource, rname)
        # get current limits
        soft, hard = resource.getrlimit(rconst)
        # check for unlimited
        if limit is None:
            limit = resource.RLIM_INFINITY
            if hard != resource.RLIM_INFINITY:
                raise StimelaProcessRuntimeError(f"can't set opts.rlimits.{name}=unlimited: hard limit is {hard}")
        else:
            if limit > hard:
                raise StimelaProcessRuntimeError(f"can't set opts.rlimits.{name}={limit}: hard limit is {hard}")
        resource.setrlimit(rconst, (limit, hard))
        log.debug(f"setting soft limit {name}={limit} (hard limit is {hard})")


def run(cab: Cab, params: Dict[str, Any], runtime: Dict[str, Any], fqname: str,
        log: logging.Logger, subst: Optional[Dict[str, Any]] = None, batch=None):
    """Runs cab contents

    Args:
        cab (Cab): cab object
        log (logger): logger to use
        subst (Optional[Dict[str, Any]]): Substitution dict for commands etc., if any.

    Returns:
        Any: return value (e.g. exit code) of content
    """
    update_rlimits(stimela.CONFIG.opts.rlimits, log)

    args = cab.flavour.get_arguments(cab, params, subst)

    log.debug(f"command line is {args}")

    cabstat = cab.reset_status()

    command_name = cab.flavour.command_name

    if batch:
        batch = SlurmBatch(**batch)
        batch.__init_cab__(cabstat.cab, params, subst, log)
        runcmd = "/bin/bash -c" + " ".join(args)
        jobfile = "foo-bar.job"
        batch.name = "foo-bar"
        batch.submit(jobfile=jobfile, runcmd=runcmd)

        return

    #-------------------------------------------------------
    # run command
    start_time = datetime.datetime.now()
    def elapsed(since=None):
        """Returns string representing elapsed time"""
        return str(datetime.datetime.now() - (since or start_time)).split('.', 1)[0]

    # log.info(f"argument lengths are {[len(a) for a in args]}")

    retcode = xrun(args[0], args[1:], shell=False, log=log,
                output_wrangler=cabstat.apply_wranglers,
                return_errcode=True, command_name=command_name, 
                gentle_ctrl_c=True,
                log_command=True if cab.flavour.log_full_command else command_name, 
                log_result=False)

    # check if output marked it as a fail
    if cabstat.success is False:
        log.error(f"declaring '{command_name}' as failed based on its output")

    # if retcode != 0 and not explicitly marked as success, mark as failed
    if retcode and cabstat.success is not True:
        cabstat.declare_failure(f"{command_name} returns error code {retcode} after {elapsed()}")
    else:
        log.info(f"{command_name} returns exit code {retcode} after {elapsed()}")

    return cabstat



import os.path, re, stat, itertools, logging, yaml, shlex
from typing import Any, List, Dict, Optional, Union
from collections import OrderedDict
from enum import Enum, IntEnum
from dataclasses import dataclass
from omegaconf import MISSING, OmegaConf
import rich.markup

from scabha.cargo import Parameter, Cargo, ListOrString, ParameterPolicies, ParameterCategory
from stimela.exceptions import CabValidationError, StimelaCabRuntimeError
from scabha.exceptions import SchemaError
from scabha.basetypes import EmptyDictDefault, EmptyListDefault
from stimela.backends import flavours
import stimela
from . import wranglers
from scabha.substitutions import substitutions_from

ParameterPassingMechanism = Enum("ParameterPassingMechanism", "args yaml", module=__name__)


@dataclass 
class CabManagement:        # defines common cab management behaviours
    environment: Optional[Dict[str, str]] = EmptyDictDefault()
    cleanup: Optional[Dict[str, ListOrString]]     = EmptyDictDefault()   
    wranglers: Optional[Dict[str, ListOrString]]   = EmptyDictDefault()   


@dataclass
class Cab(Cargo):
    """Represents a cab i.e. an atomic task in a recipe.
    See dataclass fields below for documentation of fields.

    Additional attributes available after validation with arguments:

        self.input_output:      combined parameter dict (self.input + self.output), maps name to Parameter
        self.missing_params:    dict (name to Parameter) of required parameters that have not been specified
    
    Raises:
        CabValidationError: [description]
    """
    # if set, the cab is run in a container, and this is the image name
    # if not set, commands are run by the native runner
    image: Optional[str] = None                   

    # command to run, inside the container or natively
    command: str = MISSING

    # if set, activates this virtual environment first before running the command (not much sense doing this inside the container)
    virtual_env: Optional[str] = None

    # cab flavour. Default will run the command as a binary (inside image or virtual_env). Otherwise specify
    # a string flavour, or a mapping to specify options (see backends.flavours)
    flavour: Optional[Any] = None

    # controls how params are passed. args: via command line argument, yml: via a single yml string
    parameter_passing: ParameterPassingMechanism = ParameterPassingMechanism.args

    # cab management and cleanup definitions
    management: CabManagement = CabManagement()

    # default parameter conversion policies
    policies: ParameterPolicies = ParameterPolicies()

    # For callable-type cabs, determines how the return value is treated.
    # None to ignore, "{}" to treat it as a dict of outputs, else an output name to 
    # treat it as a single output
    return_outputs: Optional[str] = "{}" 

    # runtime settings
    backend: Optional['stimela.config.Backend']
    runtime: Dict[str, Any] = EmptyDictDefault()

    _path: Optional[str] = None   # path to image definition yaml file, if any

    def __post_init__ (self):
        if self.name is None:
            self.name = self.image or self.command.split()[0]
        Cargo.__post_init__(self)
        for param in self.inputs.keys():
            if param in self.outputs:
                raise CabValidationError(f"cab {self.name}: parameter '{param}' appears in both inputs and outputs")

        # setup wranglers
        self._wranglers = []
        for pattern, actions in self.management.wranglers.items():
            self._wranglers.append(wranglers.create_list(pattern, actions))

        # check flavours
        self.flavour = flavours.init_cab_flavour(self)


    def summary(self, params=None, recursive=True, ignore_missing=False):
        lines = [f"cab {self.name}:"] 
        if params is not None:
            for name, value in params.items():
                # if type(value) is validate.Error:
                #     lines.append(f"  {name} = ERR: {value}")
                # else:
                lines.append(f"  {name} = {value}")
            lines += [f"  {name} = ???" for name, schema in self.inputs_outputs.items()
                        if name not in params and (not ignore_missing or schema.required)]
        return lines

    def rich_help(self, tree, max_category=ParameterCategory.Optional):
        tree.add(f"command: {self.command}")
        if self.image:
            tree.add(f"image: {self.image}")
        if self.virtual_env:
            tree.add(f"virtual environment: {self.virtual_env}")
        Cargo.rich_help(self, tree, max_category=max_category)

    def get_schema_policy(self, schema, policy, default=None):
        """Resolves a policy setting. If the policy is set here, returns it. If None and set in the cab,
        returns that. Else returns default value.
        """
        if getattr(schema.policies, policy) is not None:
            return getattr(schema.policies, policy)
        elif getattr(self.policies, policy) is not None:
            return getattr(self.policies, policy)
        else:
            return default

    def build_command_line(self, params: Dict[str, Any], subst: Optional[Dict[str, Any]] = None, search=True):

        try:
            with substitutions_from(subst, raise_errors=True) as context:
                venv = context.evaluate(self.virtual_env, location=["virtual_env"])
                command = context.evaluate(self.command, location=["command"])
        except Exception as exc:
            raise CabValidationError(f"error constructing cab command", exc)

        if venv:
            venv = os.path.expanduser(venv)
            if not os.path.isfile(f"{venv}/bin/activate"):
                raise CabValidationError(f"virtual environment {venv} doesn't exist")
            self.log.debug(f"virtual environment is {venv}")
        else:
            venv = None

        command_line = shlex.split(os.path.expanduser(command))
        command = command_line[0]
        args = command_line[1:]
        # collect command
        if search:
            if "/" not in command:
                from scabha.proc_utils import which
                command0 = command
                command = which(command, extra_paths=venv and [f"{venv}/bin"])
                if command is None:
                    raise CabValidationError(f"{command0}: not found", log=self.log)
            else:
                if not os.path.isfile(command) or not os.stat(command).st_mode & stat.S_IXUSR:
                    raise CabValidationError(f"{command} doesn't exist or is not executable")

        self.log.debug(f"command is {command}")

        return ([command] + args + self.build_argument_list(params)), venv

    def update_environment(self, subst):
        try:
            with substitutions_from(subst, raise_errors=True) as context:
                environ = CabManagement().environment
                for key,val in self.management.environment.items():
                    environ[key] = context.evaluate(val,
                                location=["management", "environment"])
        except Exception as exc:
            raise CabValidationError(f"Error applying environment variables", exc)

        if environ:
            self.management.environment = environ


    def filter_input_params(self, params: Dict[str, Any], apply_nom_de_guerre=True):
        """Filters dict of params, returning only those that should be passed to a cab
        (i.e. inputs or named outputs, and not skipped)
        """
        filtered_params = OrderedDict()
        for name, schema in self.inputs_outputs.items():
            output_name = (apply_nom_de_guerre and schema.nom_de_guerre) or name
            if not self.get_schema_policy(schema, 'skip'):
                if schema.is_input or schema.is_named_output:
                    if name in params:
                        filtered_params[output_name] = params[name]
                    elif self.get_schema_policy(schema, 'pass_missing_as_none'):
                        filtered_params[output_name] = None
        return filtered_params


    def build_argument_list(self, params: Dict[str, Any]):
        """
        Converts command, and current dict of parameters, into a list of command-line arguments.

        pardict:     dict of parameters. If None, pulled from default config.
        positional:  list of positional parameters, if any
        mandatory:   list of mandatory parameters.
        repeat:      How to treat iterable parameter values. If a string (e.g. ","), list values will be passed as one
                    command-line argument, joined by that separator. If True, list values will be passed as
                    multiple repeated command-line options. If None, list values are not allowed.
        repeat_dict: Like repeat, but defines this behaviour per parameter. If supplied, then "repeat" is used
                    as the default for parameters not in repeat_dict.

        Returns list of arguments.
        """

        # collect parameters

        value_dict = dict(**params)

        if self.parameter_passing is ParameterPassingMechanism.yaml:
            return [yaml.safe_dump(value_dict)]

        def get_policy(schema: Parameter, policy: str, default=None):
            return self.get_schema_policy(schema, policy, default)

        def stringify_argument(name, value, schema, option=None):
            key_value = get_policy(schema, 'key_value')

            if value is None:
                return None
            if schema.dtype == "bool" and not value and get_policy(schema, 'explicit_false') is None:
                return None

            is_list = hasattr(value, '__iter__') and type(value) is not str
            format_policy = get_policy(schema, 'format')
            format_list_policy = get_policy(schema, 'format_list')
            format_scalar_policy = get_policy(schema, 'format_list_scalar')
            split_policy = get_policy(schema, 'split')
            
            if type(value) is str and split_policy:
                value = value.split(split_policy or None)
                is_list = True

            if is_list:
                # apply formatting policies to a list of values
                if format_list_policy:
                    if len(format_list_policy) != len(value):
                        raise CabValidationError("length of format_list_policy does not match length of '{name}'", log=self.log)
                    value = [fmt.format(*value, **value_dict) for fmt in format_list_policy]
                elif format_policy:
                    value = [format_policy.format(x, **value_dict) for x in value]
                else:
                    value = [str(x) for x in value]
            else:
                # apply formatting policies to a scalar valye
                if format_scalar_policy:
                    value = [fmt.format(value, **value_dict) for fmt in format_scalar_policy]
                    is_list = True
                elif format_policy:
                    value = format_policy.format(value, **value_dict)
                else:
                    value = str(value)

            if is_list:
                # check repeat policy and form up representation
                repeat_policy = get_policy(schema, 'repeat')
                if repeat_policy == "list":
                    if key_value:
                        raise CabValidationError(f"Repeat policy 'list' is incompatible with schema policy 'key_value' for parameter '{name}'")
                    return [option] + list(value) if option else list(value)
                elif repeat_policy == "[]":
                    val = "[" + ",".join(value) + "]"
                    return [option] + [val] if option else val
                elif repeat_policy == "repeat":
                    return list(itertools.chain([option, x] for x in value)) if option else list(value)
                elif type(repeat_policy) is str:
                    return [option, repeat_policy.join(value)] if option else repeat_policy.join(value)
                elif repeat_policy is None:
                    raise CabValidationError(f"list-type parameter '{name}' does not have a repeat policy set", log=self.log)
                else:
                    raise SchemaError(f"unknown repeat policy '{repeat_policy}'", log=self.log)
            else:
                return [option, value] if option else [value]

        # check for missing parameters and collect positionals

        pos_args = [], []

        for name, schema in self.inputs_outputs.items():
            if schema.required and name not in value_dict:
                raise CabValidationError(f"required parameter '{name}' is missing", log=self.log)
            if schema.is_output and not schema.is_named_output:
                continue
            if name in value_dict:
                positional_first = get_policy(schema, 'positional_head') 
                positional = get_policy(schema, 'positional') or positional_first
                skip = get_policy(schema, 'skip') or (schema.implicit and get_policy(schema, 'skip_implicits', True))
                if positional:
                    if not skip:
                        pargs = pos_args[0 if positional_first else 1]
                        value = stringify_argument(name, value_dict[name], schema)
                        if type(value) is list:
                            pargs += value
                        elif value is not None:
                            pargs.append(value)
                    value_dict.pop(name)

        args = []
                    
        # now check for optional parameters that remain in the dict
        for name, value in value_dict.items():
            if name not in self.inputs_outputs:
                raise RuntimeError(f"unknown parameter '{name}'")
            schema = self.inputs_outputs[name]
            if schema.is_output and not schema.is_named_output:
                continue

            # default behaviour for unset skip_implicits is True
            skip_implicits = get_policy(schema, 'skip_implicits', True)

            if get_policy(schema, 'skip') or (schema.implicit and skip_implicits):
                continue

            key_value = get_policy(schema, 'key_value')

            # apply replacementss
            replacements = get_policy(schema, 'replace')
            if replacements:
                for rep_from, rep_to in replacements.items():
                    try:
                        name = name.replace(rep_from, rep_to)
                    except TypeError:
                        raise TypeError(f"Could not perform policy replacement for parameter [{name}] : {rep_from} => {rep_to}")

            prefix = get_policy(schema, 'prefix')
            if prefix is None:
                prefix = "--"
            option = prefix + (schema.nom_de_guerre or name)

            if schema.dtype == "bool":
                explicit = get_policy(schema, "explicit_" + str(value).lower()) 
                # if explicit setting is given, this also becomes the option value
                # in key=value mode, just give that value directly
                strval = str(value) if explicit is None else str(explicit)
                if key_value:
                    args += [f"{option}={strval}"]
                # in option mode, use --option value for explicit settings, 
                # else give option for True and omit for False. TODO: some tools may eventually need a policy for
                # passing --no-option for False.
                else:
                    args += [option, strval] if explicit is not None else ([option] if value else [])
            else:
                value = stringify_argument(name, value, schema, option=option)
                if type(value) is list:
                    if key_value:
                        assert len(value) == 2
                        value = [f"{value[0]}={value[1]}"]
                    args += value
                elif value is not None:
                    args.append(value)

        return pos_args[0] + args + pos_args[1]

    def reset_status(self, extra_wranglers: List = []):
        return Cab.RuntimeStatus(self, extra_wranglers=extra_wranglers)

    class RuntimeStatus(object):
        """Represents the runtime status of a cab"""

        def __init__(self, cab: "Cab", extra_wranglers: List = []):
            self.cab = cab
            self.wranglers = list(cab._wranglers) + list(extra_wranglers)
            self._success = None
            self._errors = []
            self._warnings = []
            self._outputs = OrderedDict()

        @property
        def success(self):
            return self._success

        @property
        def errors(self):
            return self._errors

        @property
        def warnings(self):
            return self._warnings

        @property
        def outputs(self):
            return self._outputs

        def declare_success(self):
            if self._success is None:
                self._success = True

        def declare_failure(self, error: Optional[Union[str, Exception]] = None):
            self._success = False
            if error is not None:
                if type(error) is str:
                    error = StimelaCabRuntimeError(error)
                self._errors.append(error)

        def declare_warning(self, message: str):
            self._warnings.append(message)

        def declare_outputs(self, outputs: Dict):
            self._outputs.update(**outputs)

        def apply_wranglers(self, output, severity):
            # make sure any unintended [rich style] tags are escaped in output
            output = rich.markup.escape(output)
            suppress = False
            for regex, wranglers in self.wranglers:
                match = regex.search(output) 
                if match:
                    for wrangler in wranglers:
                        mod_output, mod_severity = wrangler.apply(self, output, match)
                        # has wrangler asked to suppress the output?
                        if mod_output is None:
                            suppress = True
                        else:
                            output = mod_output
                        # has wrangler modified the severity?
                        if mod_severity is not None:
                            severity = max(severity, mod_severity)

            return (None, 0) if suppress else (output, severity)

CabSchema = None

def get_cab_schema():
    global CabSchema
    if CabSchema is None:
        import stimela.config
        CabSchema = OmegaConf.structured(Cab)
    return CabSchema



import click
from typing import List
from stimela.main import cli
import stimela


@cli.command(
    help="""(Re)build stimela base images. Specify a list of image names and versions (if a version is omitted, builds all known versions of image),
            or else use --all to build all required base images. 
         """,
    short_help="build stimela base images",
    no_args_is_help=True)
@click.argument("images", nargs=-1, metavar="NAME[:VERSION]", required=False) 
#                help="image/version to build (builds all versions of image by default)")
@click.option("-a", "--all", is_flag=True, 
                help="build all unavailable images (all images, in combination with --rebuild)")
@click.option("-r", "--rebuild", is_flag=True, 
                help="force rebuild of image(s)")
def build(images: List[str], all=False, rebuild=False):
    from stimela.main import BACKEND
    from stimela import CONFIG
    log = stimela.logger()

    available_images = BACKEND.available_images()
    if all:
        build_images = CONFIG.base.keys()
    else:
        build_images = images

    if not build_images:
        log.info("No images specified. Run 'stimela build -h' for help.")
        return 0

    for imagename in build_images:
        if ':' in imagename:
            imagename, version = imagename.split(":", 1)
        else:
            version = None

        if imagename not in CONFIG.base:
            log.error(f"base image '{imagename}' is not known to Stimela")
            return 2

        image = CONFIG.base[imagename]

        if version is None:
            build_versions = image.images.keys()
        elif version in image.images:
            build_versions = [version]
        else:
            log.error(f"version '{version}' is not defined for base image '{imagename}'")
            return 2

        # now loop over build versions
        for version in build_versions:
            
            # check if already exists
            if imagename in available_images and version in available_images[imagename]:
                if not rebuild:
                    log.info(f"image '{imagename}:{version}' already exists, skipping")
                    continue

            BACKEND.build(image, version)



try:
    import kubernetes
    AVAILABLE = True
    STATUS = "ok"
    from .run_kube import run
except ImportError:
    AVAILABLE = False
    STATUS = "please reinstall with the optional kube dependency (stimela[kube])"

    def run(*args, **kw):
        raise RuntimeError(f"kubernetes backend {STATUS}")



import traceback
import os
import signal
import datetime
import asyncio
import logging
from rich.markup import escape

from stimela import stimelogging

from stimela.exceptions import StimelaCabRuntimeError, StimelaProcessRuntimeError

DEBUG = 0

log = None

def get_stimela_logger():
    """Returns Stimela's logger, or None if no Stimela installed"""
    try:
        import stimela
        return stimela.logger()
    except ImportError:
        return None


def dispatch_to_log(log, line, command_name, stream_name, output_wrangler):
    # dispatch output to log
    extra = dict()
    # severity = logging.WARNING if fobj is proc.stderr else logging.INFO
    severity = logging.INFO
    if stream_name == 'stdout':
        extra['style'] = 'dim'
        extra['prefix'] = "#"
    else:
        extra['style'] = 'white'
        extra['prefix'] = "#"
    # feed through wrangler to adjust severity and content
    if output_wrangler is not None:
        line, severity = output_wrangler(escape(line), severity)
    if line is not None:
        if severity >= logging.ERROR:
            extra['prefix'] = stimelogging.FunkyMessage("[red]:warning: [/red]", "!")
        if isinstance(line, stimelogging.FunkyMessage) and line.prefix:
            extra['prefix'] = line.prefix
        log.log(severity, line, extra=extra)



def xrun(command, options, log=None, env=None, timeout=-1, kill_callback=None, output_wrangler=None, shell=True, 
            return_errcode=False, command_name=None, progress_bar=False, 
            gentle_ctrl_c=False,
            log_command=True, log_result=True):
    
    command_name = command_name or command

    # this part could be inside the container
    command_line = " ".join([command] + list(map(str, options)))
    if shell:
        command_line = " ".join([command] + list(map(str, options)))
        command = [command_line]
    else:
        command = [command] + list(map(str, options))
        command_line = " ".join(command)

    log = log or get_stimela_logger()

    if log is None:
        from .xrun_poll import xrun_nolog
        return xrun_nolog(command, name=command_name, shell=shell)

    # this part is never inside the container
    import stimela

    log = log or stimela.logger()

    if log_command:
        if log_command is True:
            log.info(f"running {command_line}", extra=dict(prefix="###", style="dim"))
        else:
            log.info(f"running {log_command}", extra=dict(prefix="###", style="dim"))
            log.debug(f"full command line is {command_line}", extra=dict(prefix="###", style="dim"))

    with stimelogging.declare_subcommand(os.path.basename(command_name)) as command_context:

        start_time = datetime.datetime.now()
        def elapsed():
            """Returns string representing elapsed time"""
            return str(datetime.datetime.now() - start_time).split('.', 1)[0]
        
        loop = asyncio.get_event_loop()
        
        proc = loop.run_until_complete(
                asyncio.create_subprocess_exec(*command,
                    stdout=asyncio.subprocess.PIPE,
                    stderr=asyncio.subprocess.PIPE))

        async def stream_reader(stream, stream_name):
            while not stream.at_eof():
                line = await stream.readline()
                line = (line.decode('utf-8') if type(line) is bytes else line).rstrip()
                if line or not stream.at_eof():
                    dispatch_to_log(log, line, command_name, stream_name, output_wrangler=output_wrangler)

        async def proc_awaiter(proc, *cancellables):
            await proc.wait()
            for task in cancellables:
                task.cancel()

        reporter = asyncio.Task(stimelogging.run_process_status_update())
        ctrl_c_caught = job_interrupted = False
        try:
            job = asyncio.gather(
                proc_awaiter(proc, reporter),
                stream_reader(proc.stdout, "stdout"),
                stream_reader(proc.stderr, "stderr"),
                reporter
            )
            results = loop.run_until_complete(job)
            status = proc.returncode
            if log_result:
                log.info(f"{command_name} exited with code {status} after {elapsed()}")
        except SystemExit as exc:
            loop.run_until_complete(proc.wait())
        except KeyboardInterrupt:
            if callable(kill_callback):
                command_context.ctrl_c()
                log.warning(f"Ctrl+C caught after {elapsed()}, shutting down {command_name} process, please give it a few moments")
                kill_callback() 
                log.info(f"the {command_name} process was shut down successfully",
                        extra=dict(stimela_subprocess_output=(command_name, "status")))
                loop.run_until_complete(proc.wait())
            else:
                try:
                    raise KeyboardInterrupt
                    # below doesn't work -- figure out later
                    if not gentle_ctrl_c or ctrl_c_caught:
                        raise KeyboardInterrupt
                    command_context.ctrl_c()
                    ctrl_c_caught = True
                    log.warning(f"Ctrl+C caught after {elapsed()}, job will error out when {command_name} process {proc.pid} completes")
                    log.warning(f"Use Ctrl+C again to interrupt the job")
                    results = loop.run_until_complete(job)
                except KeyboardInterrupt:
                    log.warning(f"Ctrl+C caught after {elapsed()}, interrupting {command_name} process {proc.pid}")
                    job_interrupted = True
                    proc.send_signal(signal.SIGINT)

                    async def wait_on_process(proc):
                        for retry in range(10):
                            await asyncio.sleep(1)
                            if proc.returncode is not None:
                                log.info(f"Process {proc.pid} has exited with return code {proc.returncode}")
                                break
                            if retry == 5:
                                log.warning(f"Process {proc.pid} not exited after {retry} seconds, will try to terminate it")
                                proc.terminate()
                            else:
                                log.info(f"Process {proc.pid} not exited after {retry} seconds, waiting a bit longer...")
                        else:
                            log.warning(f"Killing process {proc.pid}")
                            proc.kill()
                    
                    loop.run_until_complete(wait_on_process(proc))
            if job_interrupted:
                raise StimelaCabRuntimeError(f"{command_name} interrupted with Ctrl+C")
            else:
                raise StimelaCabRuntimeError(f"{command_name} complete, but received a Ctrl+C during the run")

        except Exception as exc:
            loop.run_until_complete(proc.wait())
            traceback.print_exc()
            raise StimelaCabRuntimeError(f"{command_name} threw exception: {exc} after {elapsed()}'", log=log)

        if status and not return_errcode:
            raise StimelaCabRuntimeError(f"{command_name} returns error code {status} after {elapsed()}")
    
    return status
    



import subprocess
import os
from stimela import utils
import stimela
import hashlib
from shutil import which

binary = which("singularity")
if binary:
    __version_string = subprocess.check_output([binary, "--version"]).decode("utf8")
    version = __version_string.strip().split()[-1]
    if version < "3.0.0":
        suffix = ".img"
    else:
        suffix = ".sif"
else:
    version = None

class SingularityError(Exception):
    pass

def pull(image, name, docker=True, directory=".", force=False):
    """ 
        pull an image
    """
    if docker:
        fp = "docker://{0:s}".format(image)
    else:
        fp = image
    if not os.path.exists(directory):
        os.mkdir(directory)

    image_path = os.path.abspath(os.path.join(directory, name))
    if os.path.exists(image_path) and not force:
        stimela.logger().info(f"Singularity image already exists at '{image_path}'. To replace it, please re-run with the 'force' option")
    else:
        utils.xrun(f"cd {directory} && singularity", ["pull", 
        	"--force" if force else "", "--name", 
         	name, fp])

    return 0

class Container(object):
    def __init__(self, image, name,
                 volumes=None,
                 logger=None,
                 time_out=-1,
                 runscript="/singularity",
                 environs=None,
                 workdir=None,
                 execdir="."):
        """
        Python wrapper to singularity tools for managing containers.
        """

        self.image = image
        self.volumes = volumes or []
        self.environs = environs or []
        self.logger = logger
        self.status = None
        self.WORKDIR = workdir
        self.RUNSCRIPT = runscript
        self.PID = os.getpid()
        self.uptime = "00:00:00"
        self.time_out = time_out
        self.execdir = execdir

        self._env = os.environ.copy()

        hashname = hashlib.md5(name.encode('utf-8')).hexdigest()[:3]
        self.name = hashname if version < "3.0.0" else name

    def add_volume(self, host, container, perm="rw", noverify=False):
        if os.path.exists(host) or noverify:
            if self.logger:
                self.logger.debug("Mounting volume [{0}] in container [{1}] at [{2}]".format(
                    host, self.name, container))
            host = os.path.abspath(host)
        else:
            raise IOError(
                "Path {0} cannot be mounted on container: File doesn't exist".format(host))

        self.volumes.append(":".join([host, container, perm]))

        return 0

    def add_environ(self, key, value):
        self.logger.debug("Adding environ varaible [{0}={1}] "\
                    "in container {2}".format(key, value, self.name))
        self.environs.append("=".join([key, value]))
        key_ = f"SINGULARITYENV_{key}"
	
        self.logger.debug(f"Setting singularity environmental variable {key_}={value} on host")
        self._env[key_] = value

        return 0

    def run(self, *args, output_wrangler=None):
        """
        Run a singularity container instance
        """

        if self.volumes:
            volumes = " --bind " + " --bind ".join(self.volumes)
        else:
            volumes = ""

        if not os.path.exists(self.image):
            self.logger.error(f"The image, {self.image}, required to run this cab does not exist."\
                    " Please run 'stimela pull --help' for help on how to download the image")
            raise SystemExit from None

        self.status = "running"
        self._print("Starting container [{0:s}]. Timeout set to {1:d}. The container ID is printed below.".format(
            self.name, self.time_out))
        
        utils.xrun(f"cd {self.execdir} && singularity", ["run", "--workdir", self.execdir, "--containall"] \

                    + list(args) + [volumes, self.image, self.RUNSCRIPT],
                    log=self.logger, timeout=self.time_out, output_wrangler=output_wrangler,
                    env=self._env, logfile=self.logfile)

        self.status = "exited"

        return 0

    def _print(self, message):
        if self.logger:
            self.logger.info(message)
        else:
            print(message)

        return 0



import click
from scabha.exceptions import SchemaError
from .cargo import Parameter
from typing import *
from .cargo import UNSET
from dataclasses import make_dataclass, field
from omegaconf import OmegaConf, MISSING
from collections.abc import MutableSet, MutableSequence, MutableMapping

def schema_to_dataclass(io: Dict[str, Parameter], class_name: str, bases=(), post_init: Optional[Callable] =None):
    """Converts a scabha schema to a dataclass.
       Each parameter in the schema will correspond to a field. Metadata of fields will contain:
        'help' for info, 'choices' for the choices field, 'parameter' for parameter name

    Args:
        io (Dict[str, Parameter]): dict of parameters
        class_name (str):           name of dataclass
        bases (tuple, optional): [description]. Base classes (passed to bases of make_dataclass())
        post_init (callable, optional): Inserts a __post__init__ method into the class if needed.

    Raises:
        SchemaError: if any parameters are malformed

    Returns:
        dataclass
    """
    field2name = {}
    fields = []
    for name, schema in io.items():
        if type(schema) is not Parameter:
            schema = Parameter(**schema)

        # sanitize name: dataclass won't take hyphens or periods
        # so replace with "_" and ensure uniqueness
        fldname = name.replace("-", "_").replace(".", "_")
        while fldname in field2name:
            fldname += "_"
        field2name[fldname] = name

        # form up metadata
        metadata = dict(help=schema.info, parameter=name)
        metadata.update(schema.metadata)
        if schema.choices:
            metadata['choices'] = schema.choices
        if schema.element_choices:
            metadata['element_choices'] = schema.element_choices
        metadata['required'] = required = schema.required

        if required and schema.default is not UNSET:
            raise SchemaError(
                f"Field '{fldname}' is required but specifies a default. "
                f"This behaviour is unsupported/ambiguous."
            )

        if required:
            fld = field(default=MISSING, metadata=metadata)
        elif isinstance(schema.default, MutableSequence):
            fld = field(default_factory=default_wrapper(list, schema.default),
                        metadata=metadata)
        elif isinstance(schema.default, MutableSet):
            fld = field(default_factory=default_wrapper(set, schema.default),
                        metadata=metadata)
        elif isinstance(schema.default, MutableMapping):
            fld = field(default_factory=default_wrapper(dict, schema.default),
                        metadata=metadata)
        elif schema.default is UNSET:
            fld = field(default=None, metadata=metadata)
        else:
            fld = field(default=schema.default, metadata=metadata)

        fields.append((fldname, schema._dtype, fld))

    namespace = None if post_init is None else dict(__post_init__=post_init)

    return make_dataclass(class_name, fields, bases=bases, namespace=namespace)


def default_wrapper(default_type, default_value):

    def default_factory():
        return default_type(default_value)

    return default_factory


def nested_schema_to_dataclass(nested: Dict[str, Dict], class_name: str, bases=(), section_bases=(), post_init_map={}):
    """Converts a nested schema (sections consisting of parameters) into a dataclass.

    Args:
        nested (Dict[str, Dict]):           schema sections
        class_name (str):                   name of dataclass created
        bases (tuple, optional):            base classes for outer class
        section_bases (tuple, optional):    base class for section classes
        post_init_map (dict, optional):     dict of specific __post_init__ methods to be attached to section classes

    Returns:
        tuple of (dataclass, dict):         resulting dataclass, plus dict of dataclasses for the sections
    """

    # make dataclass based on schema contents
    nested_structure = make_dataclass(f"_{class_name}_schemas", [(name, Dict[str, Parameter]) for name in nested.keys()])
    # turn into a structured config
    nested = OmegaConf.unsafe_merge(OmegaConf.structured(nested_structure), nested)
    fields = []

    # convert per-section schemas into dataclasses and make list of fields for outer dataclass
    for section, content in nested.items():
        dcls = schema_to_dataclass(content, f"{class_name}_{section}",
                                    bases=section_bases, post_init=post_init_map.get(section))

        fields.append((section, dcls, field(default=dcls())))

    # return the outer dataclass
    return make_dataclass(class_name, fields, bases=bases)


def clickify_parameters(schemas: Dict[str, Any]):

    decorator_chain = None
    for io in schemas.inputs, schemas.outputs:
        for name, schema in io.items():

            name = name.replace("_", "-")
            optname = f"--{name}"

            # sort out option type
            if schema.dtype == "bool":
                optname = f"{optname}/--no-{name}"
                dtype = bool
            elif schema.dtype == "str":
                dtype = str
            elif schema.dtype == "int":
                dtype = int
            elif schema.dtype == "float":
                dtype = float
            elif schema.dtype == "MS":
                dtype = click.Path(exists=True)

            # choices?
            if schema.choices:
                dtype = click.Choice(schema.choices)

            # aliases?
            optnames = [optname]
            if schema.abbreviation:
                optnames.append(f"-{schema.abbreviation}")

            if schema.default is UNSET:
                deco = click.option(*optnames, type=dtype,
                                    required=schema.required,
                                    metavar=schema.metavar,help=schema.info)
            else:
                deco = click.option(*optnames, type=dtype,
                                    default=schema.default, required=schema.required,
                                    metavar=schema.metavar,help=schema.info)

            if decorator_chain is None:
                decorator_chain = deco
            else:
                decorator_chain = lambda x,deco=deco,chain=decorator_chain: chain(deco(x))

    return decorator_chain



import os
import sys
import json
import yaml
import time
import tempfile
import inspect
import warnings
import re
import math
import codecs

from stimela.exceptions import StimelaCabRuntimeError, StimelaProcessRuntimeError

from .xrun_asyncio import xrun


import os.path
from dataclasses import dataclass
from typing import Union, Dict, Any, List

from stimela.kitchen.cab import Parameter

from stimela.exceptions import SchemaError
from scabha.basetypes import File, Directory, MS


@dataclass
class StimelaImageBuildInfo:
    stimela_version: str = ""
    user: str = ""
    date: str = ""
    host: str = ""  

@dataclass
class StimelaImageInfo:
    name: str = ""
    version: str = ""
    full_name: str = ""
    iid: str = ""
    build: Union[StimelaImageBuildInfo, None] = None


def resolve_required_mounts(params: Dict[str, Any], 
                            inputs: Dict[str, Parameter], 
                            outputs: Dict[str, Parameter],
                            prior_mounts: Dict[str, bool]):

    targets = {}

    # helper function to accumulate list of target paths to be mounted
    def add_target(path, must_exist, readwrite):
        if must_exist and not os.path.exists(path):
            raise SchemaError(f"{path} does not exist.")

        path = os.path.abspath(path)

        # if path doesn't exist, mount parent dir as read/write (file will be created in there)
        if not os.path.lexists(path):
            add_target(os.path.dirname(path), must_exist=True, readwrite=True)
        # else path is real
        else:
            # already mounted? Make sure readwrite is updated
            if path in targets:
                targets[path] = targets[path] or readwrite
            else:
                # not mounted, but is a link
                if os.path.islink(path):
                    # add destination as target
                    add_target(os.path.realpath(path), must_exist=must_exist, readwrite=readwrite)
                    # add parent dir as readonly target (to resolve the symlink)
                    add_target(os.path.dirname(path), must_exist=True, readwrite=False)
                # add to mounts
                else:
                    targets[path] = readwrite

    # go through parameters and accumulate target paths
    for name, value in params.items():
        schema = inputs.get(name) or outputs.get(name)
        if schema is None:
            raise SchemaError(f"parameter {name} not in defined inputs or outputs for this cab. This should have been caught by validation earlier!")

        dtype = schema._dtype 
        if dtype in (File, Directory, MS):
            files = [value]
        elif dtype in (List[File], List[Directory], List[MS]):
            files = value
        else:
            continue

        must_exist = schema.must_exist
        if must_exist is None:
            must_exist = name in inputs            
        readwrite = schema.writable or name in outputs

        # for symlink targets, we need to mount the parent directory
        for path in files:
            add_target(path, must_exist=must_exist, readwrite=readwrite)

    
    # now eliminate unnecessary targets (those that have a parent mount with the same read/write property)
    skip_targets = set()

    for path, readwrite in targets.items():
        parent = os.path.dirname(path)
        while parent != "/":  
            # if parent already mounted, and is as writeable as us, skip us
            if (parent in targets and targets[parent] >= readwrite) or \
                (parent in prior_mounts and prior_mounts[parent] >= readwrite):
                skip_targets.add(path)
                break
            parent = os.path.dirname(parent)

    for path in skip_targets:
        targets.pop(path)

    return targets

        

        


import os
from pathlib import Path

__author__ = """Sphesihle Makhathini, Oleg Smirnov and RATT"""
__email__ = "sphemakh@gmail.com"
__version__ = "2.0rc2"

CONFIG = None

UID = os.getuid()
GID = os.getgid()

root = os.path.dirname(__file__)

CAB_PATH = os.path.join(root, "cargo/cab")
BASE_PATH = os.path.join(root, "cargo/base")

# Set up logging infrastructure
LOG_HOME = os.path.expanduser("~/.stimela")
# make sure directory exists
Path(LOG_HOME).mkdir(exist_ok=True)
# This is is the default log file. It logs stimela images, containers and processes
LOG_FILE = "{0:s}/stimela_logfile.json".format(LOG_HOME)

from .stimelogging import logger, log_exception



import os
import logging
import glob
import sys
import click
import datetime
from dataclasses import dataclass
from omegaconf import OmegaConf
import stimela
from stimela import config, stimelogging

UID = stimela.UID
GID = stimela.GID
LOG_HOME = stimela.LOG_HOME
LOG_FILE = stimela.LOG_FILE

log = None

class RunExecGroup(click.Group):
    """ Makes the run and exec commands point to the same thing

    Args:
        click (_type_): _description_
    """
    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv

        alias = "exec"
        if cmd_name == alias :  
            return click.Group.get_command(self, ctx, "run")
        ctx.fail("Uknown command or alias")

    def resolve_command(self, ctx, args):
        # always return the full command name
        _, cmd, args = super().resolve_command(ctx, args)
        return cmd.name, cmd, args


@click.group(cls=RunExecGroup)
@click.option('--backend', '-b', type=click.Choice(config.Backend._member_names_), 
                help="Backend to use (for containerization).")
@click.option('--config', '-c', 'config_files', metavar='FILE', multiple=True,
                help="Extra user-defined config file(s) to load.")
@click.option('--set', '-s', 'config_dotlist', metavar='SECTION.VAR=VALUE', multiple=True,
                help="Extra user-defined config settings to apply.")
@click.option('--no-sys-config', is_flag=True, 
                help="Do not load standard config files.")
@click.option('-I', '--include', metavar="DIR", multiple=True, 
                help="Add directory to _include paths. Can be given multiple times.")
@click.option('--clear-cache', '-C', is_flag=True, 
                help="Reset the configuration cache. First thing to try in case of strange configuration errors.")
@click.option('--verbose', '-v', is_flag=True, help='Be extra verbose in output.')
@click.version_option(str(stimela.__version__))
def cli(backend, config_files=[], config_dotlist=[], include=[], verbose=False, no_sys_config=False, clear_cache=False):
    global log
    log = stimela.logger(loglevel=logging.DEBUG if verbose else logging.INFO)
    log.info(f"starting")        # remove this eventually, but it's handy for timing things right now

    stimela.VERBOSE = verbose
    if verbose:
        log.debug("verbose output enabled")

    # use this logger for exceptions
    import scabha.exceptions
    scabha.exceptions.set_logger(log)
    if verbose:
        scabha.exceptions.ALWAYS_REPORT_TRACEBACK = True

    import scabha.configuratt.cache
    scabha.configuratt.cache.set_cache_dir(os.path.expanduser("~/.cache/stimela-configs"))
    # clear cache if requested
    if clear_cache:
        scabha.configuratt.cache.clear_cache(log)

    # load config files
    stimela.CONFIG = config.load_config(extra_configs=config_files, extra_dotlist=config_dotlist, include_paths=include,
                                        verbose=verbose, use_sys_config=not no_sys_config)
    if stimela.CONFIG is None:
        log.error("failed to load configuration, exiting")
        sys.exit(1)

    if config.CONFIG_LOADED:
        log.info(f"loaded config from {config.CONFIG_LOADED}") 

    # enable logfiles and such
    if stimela.CONFIG.opts.log.enable:
        if verbose:
            stimela.CONFIG.opts.log.level = "DEBUG"
        # setup file logging
        subst = OmegaConf.create(dict(
                    info=OmegaConf.create(dict(fqname='stimela')), 
                    config=stimela.CONFIG))
        stimelogging.update_file_logger(log, stimela.CONFIG.opts.log, nesting=-1, subst=subst)


    # print backends
    log.info(f"available backends: {' '.join(config.AVAILABLE_BACKENDS)}")

    # get default backend module
    global BACKEND 
    if backend:
        stimela.CONFIG.opts.backend = backend

    if stimela.CONFIG.opts.backend.name not in config.AVAILABLE_BACKENDS:
        status = config.get_backend_status(stimela.CONFIG.opts.backend.name)
        log.error(f"the backend '{stimela.CONFIG.opts.backend.name}' is not available: {status}")
        sys.exit(1)

    log.info(f"default backend is {stimela.CONFIG.opts.backend.name}")

    # report dependencies
    for filename, attrs in config.CONFIG_DEPS.get_description().items():
        log.debug(f"config dependency {', '.join([filename] + attrs)}")

    # dump dependencies
    filename = os.path.join(stimelogging.get_logfile_dir(log) or '.', "stimela.config.deps")
    log.info(f"saving config dependencies to {filename}")
    config.CONFIG_DEPS.save(filename)


# import commands
from stimela.commands import run, images, build, push, save_config, help

## the ones not listed above haven't been converted to click yet. They are:
# cabs, clean, containers, kill, ps, pull



import glob
import os, os.path, time, sys, platform, traceback
from typing import Any, List, Dict, Optional, Union
from enum import Enum
from dataclasses import dataclass, field
from omegaconf.omegaconf import MISSING, OmegaConf
from omegaconf.errors import OmegaConfBaseException
from collections import OrderedDict

from yaml.error import YAMLError
import stimela
from stimela.exceptions import *
from stimela import log_exception

CONFIG_FILE = os.path.expanduser("~/.config/stimela.conf")

from scabha import configuratt
from scabha.cargo import ListOrString, EmptyDictDefault, EmptyListDefault
from stimela.kitchen.cab import CabManagement 


## schema for a stimela image

@dataclass
class ImageBuildInfo:
    info: Optional[str] = ""
    dockerfile: Optional[str] = "Dockerfile"
    production: Optional[bool] = True          # False can be used to mark test (non-production) images 


@dataclass
class StimelaImage:
    name: str = MISSING
    info: str = "image description"
    images: Dict[str, ImageBuildInfo] = MISSING
    _path: Optional[str] = None   # path to image definition yaml file, if any

    # optional library of common parameter sets
    params: Dict[str, Any] = EmptyDictDefault()

    # optional library of common management settings
    management: Dict[str, CabManagement] = EmptyDictDefault()


@dataclass 
class StimelaLogConfig(object):
    enable: bool = True                          
    name: str = "log-{info.fqname}"          # Default name for log file. info dict and {config.x.y} is substituted.
    ext: str = ".txt"                        # Default extension for log file.
 
    dir: str = "."                               # Default directory for log files

    symlink: Optional[str] = None                # Will make named symlink to the log directory. A useful pattern is e.g. dir="logs-{config.run.datetime}", symlink="logs",
                                                 # then each run has its own log dir, and "logs" always points to the latest one

    # how deep to nest individual log files. 0 means one log per recipe, 1 means one per step, 2 per each substep, etc. 
    nest: int = 999                             
    
    level: str = "INFO"                          # level at which we log



## overall Stimela config schema

import stimela.backends

Backend = Enum("Backend", "docker singularity podman kubernetes native", module=__name__)

SUPPORTED_BACKENDS = set(Backend.__members__)
AVAILABLE_BACKENDS = {}
BACKEND_STATUS = {}

for _name in SUPPORTED_BACKENDS:
    backend = __import__(f"stimela.backends.{_name}", fromlist=[_name])
    if getattr(backend, 'AVAILABLE', None):
        AVAILABLE_BACKENDS[_name] = backend
        BACKEND_STATUS[_name] = backend.STATUS
    else:
        if hasattr(backend, 'STATUS'):
            BACKEND_STATUS[_name] = backend.STATUS
        else:
            BACKEND_STATUS[_name] = "not implemented"

def get_backend_status(name):
    return BACKEND_STATUS.get(name, "unknown status")


@dataclass
class StimelaProfilingOptions(object):
    print_depth: int = 9999
    unroll_loops: bool = False

@dataclass
class StimelaOptions(object):
    backend: Backend = "native" #TODO(Sphe):: Maybe docker/singularity makes more sense
    registry: str = "quay.io"
    basename: str = "stimela/v2-"
    singularity_image_dir: str = "~/.singularity"
    log: StimelaLogConfig = StimelaLogConfig()
    ## list of paths to search with _include
    include: List[str] = EmptyListDefault()
    ## For distributed computes and cpu allocation
    dist: Dict[str, Any] = EmptyDictDefault()  
    ## Resource limits -- see resource module
    rlimits: Dict[str, Any] = EmptyDictDefault()
    ## Miscellaneous runtime option 
    runtime: Dict[str, Any] = EmptyDictDefault()
    ## Profiling options
    profile: StimelaProfilingOptions = StimelaProfilingOptions()


def DefaultDirs():
    return field(default_factory=lambda:dict(indir='.', outdir='.'))

_CONFIG_BASENAME = "stimela.conf"
_STIMELA_CONFDIR = os.path.os.path.expanduser("~/.stimela")

# dict of config file locations to check, in order of preference
CONFIG_LOCATIONS = OrderedDict(
    local   = _CONFIG_BASENAME,
    venv    = os.environ.get('VIRTUAL_ENV', None) and os.path.join(os.environ['VIRTUAL_ENV'], _CONFIG_BASENAME),
    stimela = os.path.isdir(_STIMELA_CONFDIR) and os.path.join(_STIMELA_CONFDIR, _CONFIG_BASENAME),
    user    = os.path.join(os.path.expanduser("~/.config"), _CONFIG_BASENAME),
)

# set to the config file that was actually found
CONFIG_LOADED = None

# set to the set of config dependencies
CONFIG_DEPS = None

# stimela base directory
STIMELA_DIR = os.path.dirname(stimela.__file__)


def merge_extra_config(conf, newconf):
    from stimela import logger

    if 'cabs' in newconf:
        for cab in newconf.cabs:
            if cab in conf.cabs:
                logger().warning(f"changing definition of cab '{cab}'")
    return OmegaConf.unsafe_merge(conf, newconf)


StimelaConfigSchema = None

ConfigExceptionTypes = (configuratt.ConfigurattError, OmegaConfBaseException, YAMLError)

def get_initial_deps():
    dependencies = configuratt.ConfigDependencies()

    # add ourselves to dependencies
    dependencies.add(STIMELA_DIR, version=stimela.__version__)              
    dependencies.add(__file__, origin=STIMELA_DIR)              
    dependencies.add(configuratt.__file__, origin=STIMELA_DIR)  

    return dependencies


def load_config(extra_configs: List[str], extra_dotlist: List[str] = [], include_paths: List[str] = [],
                verbose: bool = False, use_sys_config: bool = True):

    # # disable OmegaConf resolvers
    # for name in "oc.create", "oc.decode", "oc.deprecated", "oc.env", "oc.select", "oc.dict.keys", "oc.dict.values":
    #     print(f"clearing {name}")
    #     print(OmegaConf.clear_resolver(name))

    log = stimela.logger()

    configuratt.common.PACKAGE_VERSION = f"stimela=={stimela.__version__}"
    # set up include paths

    # stadard system paths
    configuratt.PATH[0:0] = [os.path.expanduser("~/lib/stimela"), "/usr/lib/stimela", "/usr/local/lib/stimela"]
    if 'VIRTUAL_ENV' in os.environ:
        configuratt.PATH.insert(0, os.environ['VIRTUAL_ENV'])
        configuratt.PATH.insert(0, os.path.join(os.environ['VIRTUAL_ENV'], "lib/stimela"))
    if os.path.isdir(_STIMELA_CONFDIR):
        configuratt.PATH.insert(0, _STIMELA_CONFDIR)

    # add paths from command line and environment variable
    paths = [os.path.expanduser(path) for path in include_paths]
    envpaths = os.environ.get("STIMELA_INCLUDE")
    if envpaths:
        paths += envpaths.split(':')

    if paths:
        log.info(f"added include paths: {' '.join(paths)}")
        configuratt.PATH[0:0] = paths

    if verbose:
        log.info(f"include paths are {':'.join(configuratt.PATH)}")

    extra_cache_keys = list(extra_dotlist) + configuratt.PATH

    STIMELA_DIR = os.path.dirname(stimela.__file__)
    from stimela.kitchen.cab import Cab

    global StimelaConfigSchema, StimelaLibrary, StimelaConfig
    @dataclass
    class StimelaLibrary(object):
        params: Dict[str, Any] = EmptyDictDefault()
        recipes: Dict[str, Any] = EmptyDictDefault()
        steps: Dict[str, Any] = EmptyDictDefault()
        misc: Dict[str, Any] = EmptyDictDefault()
        wisdom: Dict[str, Any] = EmptyDictDefault()

    @dataclass 
    class StimelaConfig:
        base: Dict[str, StimelaImage] = EmptyDictDefault()
        lib: StimelaLibrary = StimelaLibrary()
        cabs: Dict[str, Cab] = EmptyDictDefault()
        opts: StimelaOptions = StimelaOptions()
        vars: Dict[str, Any] = EmptyDictDefault()
        run:  Dict[str, Any] = EmptyDictDefault()

    base_configs_glob = f"{STIMELA_DIR}/cargo/base/*/*.yaml"
    lib_configs_glob = f"{STIMELA_DIR}/cargo/lib/params/*.yaml"
    cab_configs_glob = f"{STIMELA_DIR}/cargo/cab/*.yaml"

    base_configs = glob.glob(base_configs_glob)
    lib_configs = glob.glob(lib_configs_glob)
    cab_configs = glob.glob(cab_configs_glob)

    if use_sys_config:
        sys_configs = [config_file for config_file in CONFIG_LOCATIONS.values()
                        if config_file and os.path.exists(config_file)]
    else:
        sys_configs = []

    all_configs = base_configs + lib_configs + cab_configs + sys_configs + list(extra_configs)

    conf, dependencies = configuratt.load_cache(all_configs, extra_keys=extra_cache_keys, verbose=verbose) 

    if conf is not None:
        log.info("loaded full configuration from cache")
    else:
        log.info("loading configuration")
        dependencies = get_initial_deps()

        # start with empty structured config containing schema
        base_schema = OmegaConf.structured(StimelaImage) 
        cab_schema = OmegaConf.structured(Cab)
        opts_schema = OmegaConf.structured(StimelaOptions)

        StimelaConfigSchema = OmegaConf.structured(StimelaConfig)

        conf = StimelaConfigSchema.copy()

        # merge base/*/*yaml files into the config, under base.imagename
        try:
            conf.base, deps = configuratt.load_nested(base_configs, use_sources=[conf], structured=base_schema, 
                                                        nameattr='name', include_path='_path', location='base', 
                                                        use_cache=False, verbose=verbose)
            dependencies.update(deps)
        except Exception as exc:
            log_exception(ConfigError(f"error loading base configuration", exc))
            return None

        # merge lib/params/*yaml files into the config
        try:
            conf.lib.params, deps = configuratt.load_nested(lib_configs,  use_sources=[conf], 
                                                                location='lib.params', include_path='_path') 
            dependencies.update(deps)
        except Exception as exc:
            if verbose:
                traceback.print_exc()
            log_exception(ConfigError("error loading lib.params configuration", exc))
            return None

        # merge all cab/*/*yaml files into the config, under cab.taskname
        try:
            conf.cabs, deps = configuratt.load_nested(cab_configs, use_sources=[conf], structured=cab_schema, 
                                                        nameattr='name', include_path='_path', location='cabs', 
                                                        use_cache=False, verbose=verbose)
            dependencies.update(deps)
        except Exception as exc:
            if verbose:
                traceback.print_exc()
            log_exception(ConfigError("error loading cabs configuration", exc))
            return None

        conf.opts = opts_schema

        def _load(conf, config_file):
            global CONFIG_LOADED
            log.info(f"loading config from {config_file}")
            try:
                newconf, deps = configuratt.load(config_file, use_sources=[conf], verbose=verbose, use_cache=False)
                dependencies.update(deps)
                conf = merge_extra_config(conf, newconf)
                if not CONFIG_LOADED:
                    CONFIG_LOADED = config_file
            except ConfigExceptionTypes as exc:
                if verbose:
                    traceback.print_exc()
                log_exception(ConfigError("error reading {config_file}", exc))
            return conf

        # add standard configs 
        for config_file in sys_configs:
            conf = _load(conf, config_file)

        # add local configs
        for path in extra_configs:
            conf = _load(conf, path)

        if not CONFIG_LOADED:
            log.info("no user-supplied configuration files given, using defaults")

        dependencies.replace((base_configs_glob, cab_configs_glob, lib_configs_glob), STIMELA_DIR)

        configuratt.save_cache(all_configs, conf, dependencies, extra_keys=extra_cache_keys, verbose=verbose)

    # add dotlist settings
    if extra_dotlist:
        try:
            dotlist_conf = OmegaConf.from_dotlist(extra_dotlist)
            conf = OmegaConf.unsafe_merge(conf, dotlist_conf)
        except Exception as exc:
            if verbose:
                traceback.print_exc()
            log_exception(f"error applying command-line config settings", exc)
            return None

    # add runtime info
    _ds = time.strftime("%Y%m%d")
    _ts = time.strftime("%H%M%S")
    runtime = dict(
        date=_ds, 
        time=_ts, datetime=f"{_ds}-{_ts}", 
        node=platform.node().split('.', 1)[0],
        hostname=platform.node(), 
        env={key: value.replace('${', '\${') for key, value in os.environ.items()})

    conf.run = OmegaConf.create(runtime)

    # add include paths
    if conf.opts.include:
        configuratt.PATH += list(conf.opts.include)
        log.info(f"added include paths: {' '.join(conf.opts.include)}")

    global CONFIG_DEPS
    CONFIG_DEPS = dependencies

    # check for missing requirements
    missing = configuratt.check_requirements(conf, [], strict=True)
    for (loc, name, _) in missing:
        log.warn(f"optional config section '{loc}' omitted due to unmet requirement '{name}'")

    return conf




import pytest
import traceback
from scabha.exceptions import SubstitutionError
from scabha.substitutions import *
from omegaconf import OmegaConf 
from scabha.evaluator import UNSET, Evaluator
from scabha.validate import Unresolved

def test_subst():

    x = OmegaConf.create()
    x.a = 1
    x.b = "{foo.a} not meant to be substituted here since x marked as not mutable"
    x.c = 3

    ns = SubstitutionNS(foo={})

    bar = SubstitutionNS()
    ns._add_("x", x, nosubst=True)
    ns._add_("bar", bar)
    
    ns.foo.zero = 0

    ns.foo.a = "{x.a}-{x.c}"
    ns.foo.b = "{foo.a}{{}}"
    ns.foo.c = "{bar.a}-{bar.x}-{bar.b}"
#    ns.foo['d/e'] = "x"

    ns.bar.a = 1
    ns.bar.b = "{foo.b}"
    ns.bar.c = "{foo.x} deliberately unresolved"
    ns.bar.c1 = "{foo.x.y.z} deliberately unresolved"
    ns.bar.b1 = "{bar.b}"

    # some deliberate cyclics
    ns.bar.d = "{bar.d}"
    ns.bar.e = "{bar.f}"
    ns.bar.f = "{bar.e}"


    with substitutions_from(ns, raise_errors=True) as context:
        assert context.evaluate("{bar.a}") == "1"
        assert context.evaluate("{bar.b}") == "1-3{}"
        assert context.evaluate("{bar.b1}") == "1-3{}"
        assert context.evaluate(["{x.a}-{x.c}", "{foo.a}{{}}"]) == ["1-3", "1-3{}"]
        assert context.evaluate(["{x.a}-{x.c}", {'y': "{foo.a}{{}}"}]) == ["1-3", {'y': "1-3{}"}]\
        
#        print(context.evaluate("{foo.d/e}"))

    with substitutions_from(ns, raise_errors=False) as context:
        val = context.evaluate("{bar.c}")
        # expect 1 error
        assert val == ''
        assert len(context.errors) == 1
        print(f"bar.c evaluates to type{type(val)}: '{val}'")
        print(f"error is (expected): {context.errors[0]}")

    with forgiving_substitutions_from(ns) as context:
        val = context.evaluate("{nothing}")
        assert val == ''   # '{nothing}' evaluates to '' in forgiving mode

    with forgiving_substitutions_from(ns, True) as context:
        val1 = context.evaluate("{nothing}")
        val2 = context.evaluate("{nothing.more}")
        print(f"errors (none expected): {context.errors}")
        assert not context.errors
        assert val1 == "(KeyError: 'nothing')" # '{nothing}' evaluates to '(error message)' in forgiving=True mode
        assert val2 == "(KeyError: 'nothing')" # '{nothing}' evaluates to '(error message)' in forgiving=True mode

    with forgiving_substitutions_from(ns, "XX") as context: # unknown substitutions evaluate to "XX"
        val = context.evaluate("{nothing}")
        assert val == 'XX'                            
        val = context.evaluate("{bar.c}")
        assert val == 'XX deliberately unresolved'    
        val = context.evaluate("{bar.c1}")
        assert val == 'XX deliberately unresolved'    
        val = context.evaluate("{bug.x} {bug.y}")
        assert val == 'XX XX'    

    with substitutions_from(ns) as context:
        val = context.evaluate("{bar.d}")
        val = context.evaluate("{bar.e}")
        val = context.evaluate("{foo.a:02d}")
        # expect 1 error
        assert len(context.errors) == 3
        for err in context.errors:
            print(f"expected error: {err}")

    with substitutions_from(ns, raise_errors=True) as context:
        try:
            val = context.evaluate("{bar.d}")
            assert val == "not allowed"
            print("{bar.d} is ", val)
        except CyclicSubstitutionError as exc:
            traceback.print_exc()

    try:
        context.evaluate("xxx")
        raise RuntimeError("exception should have been raised due to invalid substitution")
    except SubstitutionError as exc:
        print(f"Error as expected ({exc})")


            
        
def test_formulas():

    ns = SubstitutionNS(previous={}, previous2={})

    ns.previous.x = 1
    ns.previous.x0 = 0
    ns.previous.y = 'y'
    ns.previous.z = 'z'
    ns.previous2.z = 'zz'

    current = OrderedDict(
        a = 'a{previous.x}',
        b = '==escaped',
        c = '=previous.x',
        d = '=IFSET(previous.x)',
        e = '=IFSET(previous.x,"z",2)',
        e1 = '=IFSET(previous.x,SELF,2)',
        f = '=IFSET(previous.xx)',
        g = '=IFSET(previous.xx,SELF,2)',
        h = "=IF(previous.x, True, 'False')", 
        i = "=IF(previous.x0, True, 'False')",
        j = "=IF(previous.xx, True, 'False', UNSET)",
        k = "=current.j",
        l = "=IFSET(current.f)",
        m = "=IF((previous.x+1)*previous.x == 2, previous.x == 1, previous.y == 0)",
        n = "=IF((-previous.x+1)*(previous.x) == 0, previous.x == 1, previous.y < 0)",
        o = '=previous.z',
        p = 'x{previous.zz}',
        q = '=LIST(current.a, current.b, current.c + 1, 0)',
        r = '=not IFSET(current.a)',
        s = '=current.c + current.c + current.c',
        t = '=previous*.z'
    )
    ns._add_("current", current)
    
    with substitutions_from(ns, raise_errors=True) as context:
        evaltor = Evaluator(ns, context, location=['top'])

        results = evaltor.evaluate_dict(current, corresponding_ns=ns.current, raise_substitution_errors=False, verbose=True)

        r = results

        assert r['a'] == "a1"
        assert r['b'] == "=escaped"
        assert r['c'] == 1
        assert r['d'] == 1 
        assert r['e'] == "z"
        assert r['e1'] == 1
        assert 'f' not in r 
        assert r['g'] == 2
        assert r['h'] == True
        assert r['i'] == 'False'
        assert 'j' not in r
        assert type(r['k']) is UNSET
        assert 'l' not in r
        assert r['m'] == True
        assert r['n'] == True
        assert r['o'] == 'z'
        assert type(r['p']) is Unresolved
        assert r['q'] == ["a1", "=escaped", 2, 0]
        assert r['r'] == False
        assert r['t'] == 'zz'

if __name__ == "__main__":
    test_subst()
    test_formulas()



import re, string
from collections import OrderedDict
from dataclasses import dataclass
from contextlib import contextmanager
from typing import Any, Dict, Optional, Union, List
import threading
import fnmatch

from .exceptions import Error, SubstitutionError, CyclicSubstitutionError
from .basetypes import EmptyDictDefault, Unresolved

from omegaconf import DictConfig

# thanks to https://gist.github.com/bgusach/a967e0587d6e01e889fd1d776c5f3729
def multireplace(string, replacements, ignore_case=False):
    """
    Given a string and a replacement map, it returns the replaced string.
    :param str string: string to execute replacements on
    :param dict replacements: replacement dictionary {value to find: value to replace}
    :param bool ignore_case: whether the match should be case insensitive
    :rtype: str
    """
    # If case insensitive, we need to normalize the old string so that later a replacement
    # can be found. For instance with {"HEY": "lol"} we should match and find a replacement for "hey",
    # "HEY", "hEy", etc.
    if ignore_case:
        def normalize_old(s):
            return s.lower()
        re_mode = re.IGNORECASE

    else:
        def normalize_old(s):
            return s
        re_mode = 0

    replacements = {normalize_old(key): val for key, val in replacements.items()}

    # Place longer ones first to keep shorter substrings from matching where the longer ones should take place
    # For instance given the replacements {'ab': 'AB', 'abc': 'ABC'} against the string 'hey abc', it should produce
    # 'hey ABC' and not 'hey ABc'
    rep_sorted = sorted(replacements, key=len, reverse=True)
    rep_escaped = map(re.escape, rep_sorted)

    # Create a big OR regex that matches any of the substrings to replace
    pattern = re.compile("|".join(rep_escaped), re_mode)

    # For each match, look up the new string in the replacements, being the key the normalized old string
    return pattern.sub(
        lambda match: replacements[normalize_old(match.group(0))], string
    )


class SubstitutionNS(OrderedDict):
    """Implements a namespace for {}-substitutions"""

    class StringWrapper(object):
        """Helper class used in forgiving substitutions: returns itself for all attribute access.

        This allows substitutions like "{x.y.z} xxx" to fail gracefully if "x" is missing. When the string formater
        looks up "x", it gets back a wrapper, in which it proceeds to look up "y" and "z", eventually evaluating to
        the given value that the wrapper was constructed with.
        """
        def __init__(self,  value):
            self.value = str(value)
        
        def __getitem__(self, key):
            return self

        def __getattr__(self, key):
            return self

        def __str__(self):
            return getattr(self, 'value')

    def __init__(self, **kw):
        """Initializes the namespace. Keywords are _add_'ed as items in the namespace"""
        name = kw.pop("_name_", [])
        nosubst = kw.pop("_nosubst_", False)
        super().__setattr__("_name_", name)
        super().__setattr__("_nosubst_", nosubst)
        SubstitutionNS._update_(self, **kw)

    def copy(self):
        newcopy = SubstitutionNS()
        OrderedDict.__setattr__(newcopy, "_name_", self._name_)
        OrderedDict.__setattr__(newcopy, "_nosubst_", set())
        for key, value in self.items():
            OrderedDict.__setitem__(newcopy, key, value)
        return newcopy

    def _update_(self, **kw):
        """Updates items in the namespace using _add_()"""
        for name, value in kw.items():
            SubstitutionNS._add_(self, name, value)

    def _merge_(self, ns):
        """Recursively merges in one namespace into another"""
        for name, value in ns.items():
            if name not in self:
                SubstitutionNS._add_(self, name, value)
            else:
                old_value = super().get(name)
                if isinstance(old_value, SubstitutionNS) and \
                    isinstance(value, (dict, OrderedDict, SubstitutionNS, DictConfig)):
                    old_value._merge_(value)
                else:
                    SubstitutionNS._add_(self, name, value)

    def _add_(self, name: str, value: Any, nosubst=False):
        """Adds an item to the namespace.

        Args:
            name (str): item key
            value (Any): item value. A dict or OrderedDict value becomes a SubstitutionNS automatically, with nosubst property
            nosubst (bool): use this as the nosubst property of the sub-namespace
        """
        # names with dots turned into sub-namespaces
        if '.' in name:
            subns_name, key = name.split('.', 1)
            if subns_name in self:
                subns = super().__getitem__(subns_name)
                if type(subns) is not SubstitutionNS:
                    raise TypeError(f"can't insert '{name}': {subns_name} is not a nested namespace")
            else:
                subns = SubstitutionNS(_nosubst_=nosubst or self._nosubst_, _name_=self._name_ + [subns_name])
            subns._add_(key, value)
            super().__setitem__(subns_name, subns)
        else:
            # a nested dict value becomes a substitution namespace automatically
            if type(value) in (dict, OrderedDict, DictConfig):
                value = SubstitutionNS(_nosubst_=nosubst or self._nosubst_, _name_=self._name_ + [name], **value)
            # if isinstance(value, SubstitutionNS):
            #     OrderedDict.__setattr__(v, "_props_", props)
            super().__setitem__(name, value)

    def __setattr__(self, name: str, value: Any) -> None:
        SubstitutionNS._add_(self, name, value)

    def __setitem__(self, k: str, v: Any) -> None:
        SubstitutionNS._add_(self, k, v)

    def get(self, name, default=None):
        context = SubstitutionContext.current()
        # keep track of nested lookups, if doing substitutions
        nestloc = context.nested_location if context else None
        value = None # set to None, in case exception handler is invoked before value is set
        try:
            if nestloc is not None:
                nestloc.append(name)
                # see if this location is already being substituted -- this is a cyclic substitution
                for otherloc, otherfrom in context.loc_stack[:-1]:
                    if otherloc == nestloc:
                        raise CyclicSubstitutionError(context.loc_stack[-1][1], otherfrom)
            # check wildcards, substitute last
            if name not in self and ('*' in name or '?' in name):
                names = sorted(fnmatch.filter(super().keys(), name))
                if names:
                    name = names[-1]
            # now look up again
            if name in self:
                value = super().get(name)
                if context:
                    if context.raise_errors and type(value) is Unresolved:
                        raise SubstitutionError(f"unresolved substitution for {name} ({value})")
                    if not self._nosubst_:
                        # recursive=False will invoke substitution on strings, but will return containers as is
                        value = context.evaluate(value, location=nestloc, recursive=False)
                return value
            elif default in (KeyError, AttributeError):
                raise default(name)
            else:
                return default
        # catch errors
        except Exception as exc:
            if context is not None:
                forgive = context.forgive_errors.get(type(exc))
                if type(forgive) is str:
                    return SubstitutionNS.StringWrapper(forgive.format(name=''.join(nestloc or []), value=value, target=name, exc=exc))
                elif forgive:
                    return SubstitutionNS.StringWrapper(f"({type(exc).__name__}: {exc})")
            # else re-raise exception
            raise

    def __getitem__(self, name):
        return self.get(name, KeyError)

    def __getattr__(self, name):
        return self.get(name, AttributeError)

    def _print_(self, prefix="", printfunc=print):
        for name, value in self.items():
            if name.startswith("_") or name.endswith("_"):
                continue
            if isinstance(value, SubstitutionNS):
                printfunc(f"{prefix}{name}:")
                value._print_(prefix + "  ")
            elif isinstance(value, Exception):
                printfunc(f"{prefix}{name}: ERR: {value}")
            else:
                printfunc(f"{prefix}{name}: {value}")


class SubstitutionFormatter(string.Formatter):
    def __init__(self, context: 'SubstitutionContext'):
        self.context = context
    
    def get_value(self, key, args, kwargs):
        return self.context.get_value(key, args, kwargs)


@dataclass
class SubstitutionContext(object):
    ns: Optional[SubstitutionNS]
    forgive_errors: Dict[Any, Optional[Union[str, bool]]] = EmptyDictDefault()
    raise_errors: bool = False

    def __post_init__(self):
        self.formatter = SubstitutionFormatter(self)
        # current substitution location. This is appended to with every nested attribute lookup
        self.nested_location = []
        # current location stack. A new element is inserted every time a nested substitution starts
        self.loc_stack = []
        # 
        self.enabled = True
        # list of erros and list of forgiven errors
        self.errors = []
        self.forgivens = []

    _current_contexts = {}

    @staticmethod
    def current() -> 'SubstitutionContext':
        return SubstitutionContext._current_contexts.get(threading.get_ident())

    def evaluate(self, value: Any, location: List[str] = [], recursive=True):
        """Formats value using the current substitution context

        Args:
            value (str): string to be formatted, or a container to be recursed into
            location:    list describing nested location of value, e.g. ['foo', 'bar', '1']
            recursive:   recurse into lists and dicts
        """
        if not self.enabled:
            raise SubstitutionError("substitution invoked outside of with clause")

        if not recursive and isinstance(value, (list, tuple, dict, OrderedDict)):
            return value

        # not a string, or an Error, or no "{" symbol -- return as is
        if self.ns is not None and isinstance(value, (str, list, tuple, dict, OrderedDict)):

            # an evaluate() call means a new substitution is being done. 
            # add a location list to the stack. This list will be appended to as we look up sub-attributes
            nesting = len(self.loc_stack)
            self.nested_location = []
            self.loc_stack.append((None, location))

            try:
                value = self._evaluate_element(value, location, nesting)
            finally:
                while self.loc_stack[-1][0] is not None:
                    self.loc_stack.pop()
                self.loc_stack.pop()
                self.nested_location = self.loc_stack[-1][0] if self.loc_stack else []

        return value

    def _evaluate_element(self, value: Any, location: List[str], nesting:int):
        newvalue = value
        if isinstance(value, str):
            if isinstance(value, Error) or "{" not in value:
                return value
            newvalue = self._evaluate_str(value, location, nesting)
        elif isinstance(value, (list, tuple)):
            for i, element in enumerate(value):
                newelement = self._evaluate_element(element, location + [str(i)], nesting)
                if newelement is not element:
                    if newvalue is value:
                        newvalue = list(value)
                    newvalue[i] = newelement
        elif isinstance(value, (dict, OrderedDict)):
            for key, element in value.items():
                newelement = self._evaluate_element(element, location + [key], nesting)
                if newelement is not element:
                    if newvalue is value:
                        newvalue = OrderedDict(value)
                    newvalue[key] = newelement
        return newvalue


    def _evaluate_str(self, value: str, location: List[str], nesting:int):
        try:
            # if we're doing a nested substitution, protect "{{" and "}}" from getting converted to a single brace by pre-replacing them
            if nesting:
                newvalue = multireplace(value, {"{{": "\u00AB", "}}": "\u00BB"})
            newvalue = self.formatter.format(value)
            if nesting:
                newvalue = multireplace(newvalue, {"\u00AB": "{{", "\u00BB": "}}"})
        except Exception as exc:
            # name is the object being formatted
            name = '.'.join(location)
            # target is the object that failed to be substituted in
            target = '.'.join(self.nested_location)
            # this gives us the current forgiveness policy for failed substitutions
            forgive = self.forgive_errors.get(type(exc))
            if type(forgive) is str:
                newvalue = forgive.format(name=name, value=value, target=target, exc=exc)
            elif forgive:
                newvalue = f"({exc})"
            # not forgiving error -- add to list in context, and raise if contexts wants us to raise
            else:
                locstr = f"{name}='{value}'" if name else f"'{value}'"
                if type(exc) is AttributeError:
                    err = SubstitutionError(
                        f"'{{{target}}} unresolved, in {locstr}"
                    )
                elif type(exc) is CyclicSubstitutionError:
                    err = SubstitutionError(
                        f"{{{target}}}: {exc}, in {locstr}"
                    )
                else:
                    err = SubstitutionError(
#                            f"{type(exc)} in {{{target}}}: {exc}, in {name}='{value}'"
                        f"{type(exc).__name__} at {{{target}}}: {exc}, in {locstr}"
                    )
                self.errors.append(err)
                if self.raise_errors:
                    raise
                return ''
            # dropped here, so forgive the error
            self.forgivens.append(name)
        return newvalue

            
    def get_value(self, key, args, kwargs):
        """Implements get_value for string formatter"""
        if type(key) is int:
            return args[key]
        elif key in kwargs:
            return kwargs[key]
        else:
            # this starts an attribute lookup, so remove previous lookup locations
            # a call to evaluate() will add a location of None, so rewind back to that
            while self.loc_stack[-1][0] is not None:
                self.loc_stack.pop()
            # add a new root location onto the stack
            self.nested_location = []
            self.loc_stack.append((self.nested_location, self.loc_stack[-1][1]))
            # now look up the attribute
            return self.ns.get(key, KeyError)


@contextmanager
def substitutions_from(ns: Optional[SubstitutionNS], raise_errors=False, forgive_errors: dict={}):
    thread = threading.get_ident()

    previous = SubstitutionContext._current_contexts.get(thread)
    current = SubstitutionContext(ns, forgive_errors=forgive_errors, raise_errors=raise_errors)

    SubstitutionContext._current_contexts[thread] = current

    try:
        yield current
    finally:
        SubstitutionContext._current_contexts[thread] = previous
        current.enabled = False


def forgiving_substitutions_from(ns: SubstitutionNS, forgive="", raise_errors=False):
    """"""
    forgive_errors = {err: forgive for err in [AttributeError, KeyError, TypeError, ValueError, SubstitutionError]}

    return substitutions_from(ns, raise_errors=raise_errors, forgive_errors=forgive_errors)



def perform_ll_substitutions(subst: Dict[str, Any], params: Dict[str, Any], 
                             raise_exceptions: bool=False) -> List[Exception]:
    """performs <<-substitutions on dict of parameters

    Args:
        subst (Dict[str, Any]): substitution ictionary
        params (Dict[str, Any]): parameter dictionary, modified in place
        raise_exceptions (bool, optional): raise exceptions on error. Defaults to False.

    Returns:
        List[Exception]: list of errors (if raise_exceptions is False), or [] on success
    """    

    errors = OrderedDict()
    repeat = True
    while repeat:
        repeat = False  # will reraise if need to run again (going to add that feature)
        
        for name, value in list(params.items()):
            if name not in errors and type(value) is str:
                # << is escape character -- maps back to single <
                if value.startswith("<<<"):
                    params[name] = value[1:]
                elif value.startswith("<<"):
                    target = value[2:].strip()
                    if not target:
                        exc = SubstitutionError(f"{name} = << {value}: has no <<-substitution target")
                        if raise_exceptions:
                            raise exc
                        errors[name] = exc
                        continue
                    # check for ?value :value and !value specifiers
                    components = target.split(" ")
                    true_value = false_value = undef_value = exc = None
                    if len(components) > 1:
                        target = components[0]
                        for comp in components[1:]:
                            if comp.startswith("?"):
                                true_value = comp[1:]
                            elif comp.startswith(":"):
                                false_value = comp[1:]
                            elif comp.startswith("!"):
                                undef_value = comp[1:]
                                # add trailing ? to target if it was missing
                                if '?' not in target:
                                    target = target + "?"
                            else:
                                exc = SubstitutionError(f"{name} = '{value}': invalid component '{comp}'")
                                break
                    if exc is not None:
                        if raise_exceptions:
                            raise exc
                        errors[name] = exc
                        continue
                    keys = target.split('.')
                    # look up keys in substitution dict recursively
                    # a ?-lookup is optional
                    target_value = subst
                    while keys:
                        key = keys.pop(0)
                        optional = key.endswith('?')
                        if optional:
                            key = key[:-1]
                        target_value = target_value.get(key, SubstitutionError)
                        if target_value is SubstitutionError:
                            break
                    # repeat loop eventually, since something has changed
                    repeat = True
                    # if a lookup has failed -- for optional lookups, delete the parameter, otherwise
                    # report error
                    if target_value is SubstitutionError:
                        if optional:
                            if undef_value is not None:
                                params[name] = undef_value
                            else:
                                del params[name]
                        else:
                            exc = SubstitutionError(f"{name} = '{value}': lookup failed for '{key}'")
                            if raise_exceptions:
                                raise exc
                            errors[name] = exc
                        continue
                    # success!
                    if target_value:
                        params[name] = true_value if true_value is not None else target_value
                    else:
                        params[name] = false_value if false_value is not None else target_value
                
    return list(errors.values())



import click
import stimela
from stimela.main import cli


@cli.command(
    help="""Lists all known stimela images. 
         """,
    short_help="list known stimela images")
@click.option("-i", "--print-ids", is_flag=True, 
                help="list in the more terse image+ID format.")
def images(print_ids=False):
    from stimela.main import BACKEND
    from stimela import CONFIG
    log = stimela.logger()
    available = BACKEND.available_images()

    if not print_ids:
        log.info("image list follows")

        header = f"{'IMAGE':19} {'VERSION':19} {'DESCRIPTION':19} BUILT BY"
        print(header)
        print("-"*len(header))

    for _, baseinfo in stimela.CONFIG.base.items():
        name0 = name = baseinfo.name
        for version, versinfo in baseinfo.images.items():
            if name0 in available and version in available[name0]:
                image = available[name0][version]
                status = f"{image.build.user}@{image.build.host} on {image.build.date} using stimela {image.build.stimela_version}"
            else:
                image = None
                status = "not found: please pull or build" 
            if print_ids:
                if image is None:
                    print(f"{name:19} ???")
                else:
                    iid = image.iid
                    if iid.startswith("sha256:"):
                        iid = iid.split(":", 1)[1]
                    if len(iid) > 12:
                        iid = iid[:12]
                    print(f"{name:19} {image.full_name:39} {iid}")
            else:
                print(f"{name:19} {version:19} {versinfo.info:19} {status}")
                name = ''



from typing import Any, List, Dict, Optional, Union
from dataclasses import dataclass

from scabha.basetypes import EmptyDictDefault, EmptyListDefault

from .cab import Cab


@dataclass
class Batch:
    scheduler: str = "slurm"
    cpus: int = 4
    mem: str = "128gb"
    email: Optional[str] = None

    def __init_cab__(self, cab: Cab, params: Dict[str, Any], subst: Optional[Dict[str, Any]], log: Any=None):
        self.cab = cab
        self.log = log
        self.args, self.venv = self.cab.build_command_line(params, subst)




from stimela.main import get_cab_definition
from stimela import logger, LOG_HOME, CAB_USERNAME, CAB_PATH


def make_parser(subparsers):
    parser = subparsers.add_parser("cabs", help='List executor (a.k.a cab) images')

    parser.add_argument("-l", "--list", action="store_true",
                        help="List cab names")

    parser.add_argument("-i", "--cab-doc",
                        help="Will display document about the specified cab. For example, \
to get help on the 'cleanmask cab' run 'stimela cabs --cab-doc cleanmask'")

    parser.add_argument("-ls", "--list-summary", action="store_true",
                        help="List cabs with a summary of the cab")

    parser.set_defaults(func=cabs)


def cabs(args, conf):
    pass


# ## singularity:
# # singularity inspect -l --json ubuntu.img

# ## podman:
# # 


import shlex
from typing import Optional, Any, Union, Dict
from dataclasses import dataclass

from stimela.exceptions import CabValidationError
from stimela.kitchen.cab import Cab
from scabha.cab_utils import CAB_OUTPUT_PREFIX
from stimela.kitchen import wranglers
from scabha.substitutions import substitutions_from

from . import _CallableFlavour, _BaseFlavour


@dataclass
class BinaryFlavour(_BaseFlavour):
    """
    Represents a cab flavour that is a command run via the shell
    """
    kind: str = "binary"

    def get_arguments(self, cab: Cab, params: Dict[str, Any], subst: Dict[str, Any]):

        # build command line from parameters
        args, venv = cab.build_command_line(params, subst)

        # prepend virtual env invocation, if asked
        if venv:
            args = ["/bin/bash", "--rcfile", f"{venv}/bin/activate", "-c", " ".join(shlex.quote(arg) for arg in args)]

        return args




from dataclasses import field, dataclass
from collections import OrderedDict
from typing import List
from .exceptions import UnsetError

def EmptyDictDefault():
    return field(default_factory=lambda:OrderedDict())


def EmptyListDefault():
    return field(default_factory=lambda:[])
    

@dataclass
class Unresolved(object):
    value: str = ""
    errors: List[Exception] = EmptyListDefault()

    def __post_init__(self):
        if not self.value:
            self.value = "; ".join(map(str, self.errors))
        if not self.errors:
            self.errors = [UnsetError(f"'{self.value}' undefined")]
        # prevent {}-substitutions on Unresolved message
        self.value = self.value.replace("{", "{{").replace("}", "}}")

    def __str__(self):
        return f"Unresolved({self.value})"

class UNSET(Unresolved):
    """Marks unset values in formulas"""
    pass

class Placeholder(Unresolved):
    """Marks placeholder values that are guaranteed to resolve later, such as fot-loop iterants"""
    pass

class SkippedOutput(Unresolved):
    """Marks invalid outputs of skipped steps"""
    pass

import os.path

class File(str):

    @property
    def NAME(self):
        return File(os.path.basename(self))

    @property
    def DIR(self):
        return File(os.path.dirname(self))

    @property
    def BASEPATH(self):
        return File(os.path.splitext(self)[0])

    @property
    def BASENAME(self):
        return File(os.path.splitext(self.NAME)[0])

    @property
    def EXT(self):
        return os.path.splitext(self)[1]


class Directory(File):
    pass

class MS(Directory):
    pass







from pyparsing import *
from pyparsing import common
# Forward, Group, Word, Optional, alphas, alphanums, nums, ZeroOrMore, Literal, sglQuotedString, dblQuotedString
from rich import print

def test_parser():
    from scabha.evaluator import construct_parser

    expr = construct_parser()

    for string in [
            "a.b + b.c - c.d",
            "a.b + b.c * c.d",
            "a.b + -b.c",
            "a.b", 
            "IFSET(a.b)",
                ]:
        print(f"\n\n\n=====================\nExpression: {string}\n")
        a = expr.parseString(string, parse_all=True)
        print(f"\n\n\n{a.getName()}")
        print(a.dump())

    # a = expr.parseString("IFSET(a.b)+IFSET(a.b)", parse_all=True)
    # print(a.dump())

    # a = expr.parseString("a.x==b.x", parse_all=True)
    # print(a.dump())


    # a = expr.parseString("(a.x==0)==IF(a.x==0,1,2,3)", parse_all=True)
    # print(a.dump())

    # a = expr.parseString("IFSET(a.b, (a.x==0)==(a.x==0),(a.x!=b.x))", parse_all=True)
    # print(a.dump())

    # print("===")
    # a = expr.parseString("a.b OR a.b", parse_all=True)
    # print(a.dump())

    # a = expr.parse_string("IF((previous.x+1)*previous.x == 2, previous.x == 0, previous.y == 0)", parse_all=True)

    # expr.runTests("""
    #     (a==0)
    #     ((a==0)==(a==0))
    #     IFSET(a.b)
    #     IFSET(a.b, (a==0)==(a==0),(a!=b))
    #     IF((previous.x+1)*previous.x == 2, previous.x is 0, previous.y is not 0)
    #     IF((-previous.x+1)*previous.x == 0, previous.x is 0, previous.y < 0)
    #     a.b
    #     """)

#    a = expr.parse_string("((a.x))")
    



if __name__ == "__main__":
    test_parser()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'tests', 'scabha_tests': 'tests/scabha_tests'}

packages = \
['scabha',
 'scabha.configuratt',
 'scabha_tests',
 'stimela',
 'stimela.backends',
 'stimela.backends.flavours',
 'stimela.backends.kubernetes',
 'stimela.backends.native',
 'stimela.cargo',
 'stimela.cargo.base',
 'stimela.commands',
 'stimela.kitchen',
 'stimela.schedulers',
 'stimela.utils',
 'stimela_tests']

package_data = \
{'': ['*'],
 'stimela.cargo': ['cab/*', 'lib/params/*'],
 'stimela.cargo.base': ['21cmfast/*',
                        'aegean/*',
                        'aimfast/*',
                        'aoflagger/*',
                        'astropy/*',
                        'base/*',
                        'casa/*',
                        'casarest/*',
                        'catdagger/*',
                        'chgcentre/*',
                        'codex-africanus/*',
                        'cubical/*',
                        'cubical_ddf/*',
                        'curl/*',
                        'ddfacet/*',
                        'eidos/*',
                        'katdal/*',
                        'lofar/*',
                        'meqtrees/*',
                        'montage/*',
                        'moresane/*',
                        'msutils/*',
                        'owlcat/*',
                        'politsiyakat/*',
                        'pybdsf/*',
                        'pyddi/*',
                        'quartical/*',
                        'ragavi/*',
                        'rfimasker/*',
                        'rfinder/*',
                        'sagecal/*',
                        'shadems/*',
                        'sharpener/*',
                        'sofia/*',
                        'sourcery/*',
                        'sunblocker/*',
                        'tigger/*',
                        'tricolour/*',
                        'wsclean/*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'dill>=0.3.6,<0.4.0',
 'munch>=2.5.0,<3.0.0',
 'omegaconf>=2.1,<3.0',
 'psutil>=5.9.3,<6.0.0',
 'pydantic>=1.10.2,<2.0.0',
 'pyparsing>=3.0.9,<4.0.0',
 'rich>=12.6.0,<13.0.0']

entry_points = \
{'console_scripts': ['stimela = stimela.main:cli']}

setup_kwargs = {
    'name': 'stimela',
    'version': '2.0rc2',
    'description': 'Framework for system agnostic pipelines for radio interferometry arrays',
    'long_description': '\n============\nstimela 2.0\n============\n\n\n|Pypi Version|\n|Python Versions|  \n\nA workflow management framework for radio interometry data processing pipelines.\n\n`Documentation Page <https://stimela.readthedocs.io/>`_  \n=========================================================================================\n\n\n\n.. |Pypi Version| image:: https://img.shields.io/pypi/v/stimela.svg\n                  :target: https://pypi.python.org/pypi/stimela\n                  :alt:\n\n\n.. |Python Versions| image:: https://img.shields.io/pypi/pyversions/stimela.svg\n                     :target: https://pypi.python.org/pypi/stimela\n                     :alt:\n',
    'author': 'Sphesihle Makhathini',
    'author_email': 'sphemakh@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import os.path
import importlib
import hashlib
import datetime
import fnmatch
import subprocess
from dataclasses import dataclass

from omegaconf.omegaconf import OmegaConf, DictConfig, ListConfig
from typing import Any, List, Dict, Optional, OrderedDict, Union, Callable

from .common import *

@dataclass
class FailRecord(object):
    filename: str
    origin: Optional[str]
    modulename: Optional[str]
    fname: Optional[str]

@dataclass
class RequirementRecord(object):
    location: str
    requires: str
    filename: str
    optional: bool = False


class ConfigDependencies(object):
    _git_cache = {}

    def __init__(self):
        self.deps = OmegaConf.create()
        self.fails = OmegaConf.create()
        # self.provides = OmegaConf.create()
        # self.requires = OmegaConf.create()
    
    def add(self, filename: str, origin: Optional[str]=None, missing=False, **extra_attrs):
        """Adds a file to the set of dependencies

        Args:
            filename (str): filename
            origin (str or None): if not None, marks dependency as originating from another dependency
            missing (bool, optional): If True, marks depndency as missing. Defaults to False.
        """
        filename = os.path.abspath(filename)
        if filename in self.deps:
            return
        depinfo = OmegaConf.create()
        depinfo.update(**extra_attrs)
        if origin is not None:
            depinfo.origin = origin
        else:
            if missing or not os.path.exists(filename):
                depinfo.mtime     = 0 
                depinfo.mtime_str = "n/a"
                self.deps[filename] = depinfo
                return
            # get mtime and hash
            depinfo.mtime     = os.path.getmtime(filename) 
            depinfo.mtime_str = datetime.datetime.fromtimestamp(depinfo.mtime).strftime('%c')
            if not os.path.isdir(filename):
                depinfo.md5hash   = hashlib.md5(open(filename, "rb").read()).hexdigest()
            # add git info
            dirname = os.path.realpath(filename)
            if not os.path.isdir(dirname):
                dirname = os.path.dirname(dirname) 
            gitinfo = self._get_git_info(dirname)
            if gitinfo:
                depinfo.git = gitinfo
        self.deps[filename] = depinfo

    def add_fail(self, fail: FailRecord):
        self.fails[fail.filename] = OmegaConf.structured(fail)

    def replace(self, globs: List[str], dirname: str, **extra_attrs):
        remove = set()
        for glob in globs:
            remove.update(fnmatch.filter(self.deps, glob))
        if remove:
            for name in remove:
                self.deps[name] = OmegaConf.create(dict(origin=dirname))
        # add directory
        if dirname not in self.deps:
            self.add(dirname, **extra_attrs)

    def update(self, other):
        for name in other.deps:
            if name not in self.deps:
                self.deps[name] = other.deps[name]
        self.fails = OmegaConf.unsafe_merge(self.fails, other.fails)
        # self.provides = OmegaConf.unsafe_merge(self.provides, other.provides)
        # self.requires = OmegaConf.unsafe_merge(self.requires, other.requires)

    def save(self, filename):
        OmegaConf.save(self.deps, filename)

    def _get_git_info(self, dirname: str):
        """Returns git info structure for a directory, or None if not under git control

        Args:
            dirname (str): path

        Returns:
            DictConfig: directory info
        """
        # check cache first
        if dirname in self._git_cache:
            return self._git_cache[dirname]
        try:
            branches = subprocess.check_output("git -c color.ui=never branch -a -v -v".split(), 
                                                cwd=dirname, 
                                                stderr=subprocess.DEVNULL)
        except subprocess.CalledProcessError as exc:
            self._git_cache[dirname] = None
            return None
        # use git to get the info
        gitinfo = OmegaConf.create()
        for line in branches.decode().split("\n"):
            line = line.strip()
            if line.startswith("*"):
                gitinfo.branch = line[1:].strip()
                break
        # get description
        try:
            describe = subprocess.check_output("git describe --abbrev=16 --always --long --all".split(), cwd=dirname)
            gitinfo.describe = describe.decode().strip()
        except subprocess.CalledProcessError as exc:
            pass
        # get remote info
        try:
            remotes = subprocess.check_output("git remote -v".split(), cwd=dirname)
            gitinfo.remotes = remotes.decode().strip().split('\n')
        except subprocess.CalledProcessError as exc:
            pass

        self._git_cache[dirname] = gitinfo
        return gitinfo

    def get_description(self):
        desc = OrderedDict()
        for filename, attrs in self.deps.items():
            attrs_items = attrs.items() if attrs else [] 
            attrs_str = [f"mtime: {datetime.datetime.fromtimestamp(value).strftime('%c')}" 
                            if attr == "mtime" else f"{attr}: {value}"
                            for attr, value in attrs_items]
            desc[filename] = attrs_str
        return desc

    def have_deps_changed(self, mtime, verbose=False):
        # check that all dependencies are older than the cache
        for f in self.deps.keys():
            if not os.path.exists(f):
                if verbose:
                    print(f"Dependency {f} doesn't exist, forcing reload")
                return True
            if os.path.getmtime(f) > mtime:
                if verbose:
                    print(f"Dependency {f} is newer than the cache, forcing reload")
                return True
        # check that previously failing includes are not now succeeding (because that's also reason to reload cache)
        for filename, dep in self.fails.items():
            if dep.modulename:
                try:
                    mod = importlib.import_module(dep.modulename)
                    fname = os.path.join(os.path.dirname(mod.__file__), dep.fname)
                    if os.path.exists(fname):
                        return True
                except ImportError as exc:
                    pass
            elif not dep.missing_parent:
                if os.path.exists(filename):
                    return True
        return False

    # def add_provision_record(self, loc, filename):
    #     if loc not in self.provides:
    #         self.provides[loc] = []
    #     if filename not in self.provides[loc]:
    #         self.provides[loc].append(filename)

    # def scan_requirements(self, conf: DictConfig, location: Optional[str], filename: str):
    #     # build requirements map first using recursive helper
    #     def _scan(conf, loc, filename):
    #         if isinstance(conf, DictConfig):
    #             # add to requirements map, if this has requirements
    #             for optional, keyword in (False, "_requires"), (True, "_contingent"):
    #                 reqs = pop_conf(conf, keyword, [])
    #                 reqs = [reqs] if type(reqs) is str else reqs
    #                 # make list of unmet requirements
    #                 reqs = [req for req in reqs if (loc if req == "_base" else req) not in self.provides]
    #                 # save remaining unresolved reqs
    #                 if reqs:
    #                     if loc not in self.requires:
    #                         self.requires[loc] = OmegaConf.create()
    #                     for req in reqs:
    #                         reqloc = loc if req == "_base" else req
    #                         self.requires[loc][req] = RequirementRecord(loc, reqloc, filename, optional=optional)
    #             # If all resolved, add to provision record
    #             if loc not in self.requires and loc not in self.provides:
    #                 self.provides[loc] = filename
    #             # recurse into content
    #             for name, value in conf.items_ex(resolve=False):
    #                 _scan(value, f"{loc}.{name}" if loc else name, filename)

    #     _scan(conf, location or "", filename)

    # def check_requirements(self, conf: DictConfig, strict=True):
    #     """Checks remaining unmet requirements"""
    #     unmet = []
    #     optional = {}
    #     for loc, reqs in self.requires.items_ex(resolve=False):
    #         for name, req in reqs.items_ex(resolve=False):
    #             if req.requires not in self.provides:
    #                 if not strict or req.optional:
    #                     optional[loc, name] = req
    #                 else:
    #                     unmet.append(ConfigurattError(f"requirement '{name}' not met for section '{loc}' in {req.filename}"))
    #     if unmet:
    #         raise ConfigurattError("configuration has missing requirements", nested=unmet)

    #     for loc, _ in optional.keys():
    #         section = conf
    #         loc_elems = loc.split(".")
    #         try:
    #             for loc_elem in loc_elems[:-1]:
    #                 section = section[loc_elem]
    #             del section[loc_elem[-1]]
    #         except KeyError as exc:
    #             pass
        
    #     return optional



import itertools
import click
import logging
import os.path
import yaml
import sys
import traceback

from datetime import datetime
from typing import List, Optional, Tuple
from collections import OrderedDict
from omegaconf.omegaconf import OmegaConf, OmegaConfBaseException

import stimela
from scabha import configuratt
from scabha.basetypes import UNSET
from scabha.exceptions import ScabhaBaseException
from stimela import stimelogging
import stimela.config
from stimela.config import ConfigExceptionTypes
from stimela import logger, log_exception
from stimela.exceptions import RecipeValidationError, StimelaRuntimeError, StepSelectionError
from stimela.main import cli
from stimela.kitchen.recipe import Recipe, Step, RecipeSchema, join_quote
from stimela import task_stats

def load_recipe_file(filename: str):
    dependencies = stimela.config.get_initial_deps()

    # if file contains a recipe entry, treat it as a full config (that can include cabs etc.)
    try:
        conf, deps = configuratt.load(filename, use_sources=[stimela.CONFIG], no_toplevel_cache=True)
    except ConfigExceptionTypes as exc:
        log_exception(f"error loading {filename}", exc)
        sys.exit(2)

    # warn user if any includes failed
    if deps.fails:
        logger().warning(f"{len(deps.fails)} optional includes were not found, some cabs may not be available")
        for path, dep in deps.fails.items():
            logger().warning(f"    {path} (from {dep.origin})")

    dependencies.update(deps)

    # check for missing requirements
    missing = configuratt.check_requirements(conf, [stimela.CONFIG], strict=True)
    for (loc, name, _) in missing:
        logger().warning(f"optional config section '{loc}' omitted due to missing requirement '{name}'")


    return conf, dependencies


@cli.command("run",
    help="""
    Execute a single cab, or a recipe from a YML file. 
    If the YML files contains multiple recipes, specify the recipe name as an extra argument.
    Use PARAM=VALUE to specify parameters for the recipe or cab. You can also use X.Y.Z=FOO to
    change any and all config and/or recipe settings.
    """,
    no_args_is_help=True)
@click.option("-s", "--step", "step_ranges", metavar="STEP(s)", multiple=True,
                help="""only runs specific step(s) from the recipe. Use commas, or give multiple times to cherry-pick steps.
                Use [BEGIN]:[END] to specify a range of steps. Note that cherry-picking an individual step via this option
                also impies --enable-step.""")
@click.option("-t", "--tags", "tags", metavar="TAG(s)", multiple=True,
                help="""only runs steps wth the given tags (and also steps tagged as "always"). 
                Use commas, or give multiple times for multiple tags.""")
@click.option("--skip-tags", "skip_tags", metavar="TAG(s)", multiple=True,
                help="""explicitly skips steps wth the given tags. 
                Use commas, or give multiple times for multiple tags.""")
@click.option("-e", "--enable-step", "enable_steps", metavar="STEP(s)", multiple=True,
                help="""Force-enable steps even if the recipe marks them as skipped. Use commas, or give multiple times 
                for multiple steps.""")
@click.option("-a", "--assign", metavar="PARAM VALUE", nargs=2, multiple=True,
                help="""assigns values to parameters: equivalent to PARAM=VALUE, but plays nicer with the shell's 
                tab completion.""")
@click.option("-l", "--last-recipe", is_flag=True,
                help="""if multiple recipes are defined, selects the last one for execution.""")
@click.option("-d", "--dry-run", is_flag=True,
                help="""Doesn't actually run anything, only prints the selected steps.""")
@click.option("-p", "--profile", metavar="DEPTH", type=int,
                help="""Print per-step profiling stats to this depth. 0 disables.""")
@click.argument("what", metavar="filename.yml|cab name") 
@click.argument("parameters", nargs=-1, metavar="[recipe name] [PARAM=VALUE] [X.Y.Z=FOO] ...", required=False) 
def run(what: str, parameters: List[str] = [], dry_run: bool = False, last_recipe: bool = False, profile: Optional[int] = None,
    assign: List[Tuple[str, str]] = [],
    step_ranges: List[str] = [], tags: List[str] = [], skip_tags: List[str] = [], enable_steps: List[str] = []):

    log = logger()
    params = OrderedDict()
    dotlist = OrderedDict()
    errcode = 0
    recipe_name = None

    def convert_value(value):
        if value == "=UNSET":
            return UNSET
        else:
            return yaml.safe_load(value)

    # parse assign values as YaML
    for key, value in assign:
        # parse string as yaml value
        try:
            params[key] = convert_value(value)
        except Exception as exc:
            log_exception(f"error parsing value for --assign {key} {value}", exc)
            errcode = 2

    # parse arguments as recipe name, parameter assignments, or dotlist for OmegaConf    
    for pp in parameters:
        if "=" not in pp:
            if recipe_name is not None:
                log_exception(f"multiple recipe names given")
                errcode = 2
            recipe_name = pp
        else:
            key, value = pp.split("=", 1)
            # parse string as yaml value
            try:
                params[key] = convert_value(value)
            except Exception as exc:
                log_exception(f"error parsing {pp}", exc)
                errcode = 2

    if errcode:
        sys.exit(errcode)

    # load extra config settigs from dotkey arguments, to be merged in below
    # (when loading a recipe file, we want to merge these in AFTER the recipe is loaded, because the arguments
    # might apply to the recipe)
    try:
        extra_config = OmegaConf.from_dotlist(list(dotlist.values())) if dotlist else OmegaConf.create()
    except OmegaConfBaseException as exc:
        log_exception(f"error loading command-line dotlist", exc)
        sys.exit(2)

    if what in stimela.CONFIG.cabs:
        cabname = what

        try:
            stimela.CONFIG = OmegaConf.unsafe_merge(stimela.CONFIG, extra_config)
        except OmegaConfBaseException as exc:
            log_exception(f"error applying command-line dotlist", exc)
            sys.exit(2)

        log.info(f"setting up cab {cabname}")

        # create step config by merging in settings (var=value pairs from the command line) 
        outer_step = Step(cab=cabname, params=params)

        # prevalidate() is done by run() automatically if not already done, but it does set up the recipe's logger, so do it anyway
        try:
            outer_step.prevalidate(root=True)
        except ScabhaBaseException as exc:
            log_exception(exc)
            sys.exit(1)

    else:
        if not os.path.isfile(what):
            log_exception(f"'{what}' is neither a recipe file nor a known stimela cab")
            sys.exit(2)

        log.info(f"loading recipe/config {what}")

        conf, recipe_deps = load_recipe_file(what)
        conf.merge_with(extra_config)

        # anything that is not a standard config section will be treated as a recipe
        all_recipe_names = [name for name in conf if name not in stimela.CONFIG]
        if not all_recipe_names:
            log_exception(f"{what} does not contain any recipes")
            sys.exit(2)

        # split content into config sections, and recipes:
        # config secions are merged into the config namespace, while recipes go under
        # lib.recipes
        update_conf = OmegaConf.create()
        for name, value in conf.items():
            if name in stimela.CONFIG:
                update_conf[name] = value
            else:
                try:
                    stimela.CONFIG.lib.recipes[name] = OmegaConf.merge(RecipeSchema, value)
                except Exception as exc:
                    log_exception(f"error in definition of recipe '{name}'", exc)
                    sys.exit(2)
        
        try:
            stimela.CONFIG = OmegaConf.unsafe_merge(stimela.CONFIG, update_conf)
        except Exception as exc:
            log_exception(f"error applying configuration from {what}", exc)
            sys.exit(2)

        log.info(f"{what} contains the following recipe sections: {join_quote(all_recipe_names)}")

        if recipe_name:
            if recipe_name not in conf:
                log_exception(f"{what} does not contain recipe '{recipe_name}'")
                sys.exit(2)
        elif last_recipe or len(all_recipe_names) == 1:
            recipe_name = all_recipe_names[-1]
        else:
            print(f"This file contains the following recipes: {', '.join(all_recipe_names)}")
            log_exception(f"multiple recipes found, please specify one on the command line")
            sys.exit(2)
        
        log.info(f"selected recipe is '{recipe_name}'")

        # create recipe object from the config
        kwargs = dict(**stimela.CONFIG.lib.recipes[recipe_name])
        kwargs.setdefault('name', recipe_name)
        try:
            recipe = Recipe(**kwargs)
        except Exception as exc:
            traceback.print_exc()
            log_exception(f"error loading recipe '{recipe_name}'", exc)
            sys.exit(2)

        # force name, if not set
        if not recipe.name:
            recipe.name = recipe_name

        # wrap it in an outer step and prevalidate (to set up loggers etc.)
        recipe.fqname = recipe_name
        recipe.finalize()
        
        for key, value in params.items():
            recipe.assign_value(key, value, override=True)

        # split out parameters
        params = {key: value for key, value in params.items() if key in recipe.inputs_outputs}

        stimelogging.declare_chapter("prevalidation")
        log.info("pre-validating the recipe")
        outer_step = Step(recipe=recipe, name=f"{recipe_name}", info=what, params=params)
        try:
            params = outer_step.prevalidate(root=True)
        except Exception as exc:
            log_exception(RecipeValidationError(f"pre-validation of recipe '{recipe_name}' failed", exc))
            for line in traceback.format_exc().split("\n"):
                log.debug(line)
            sys.exit(1)        

        # select recipe substeps based on command line, and exit if nothing to run
        selection_options = []
        for opts in (tags, skip_tags, step_ranges, enable_steps):
            selection_options.append(set(itertools.chain(*(opt.split(",") for opt in opts))))
        
        try:
            if not recipe.restrict_steps(*selection_options):
                sys.exit(0)
        except StepSelectionError as exc:
            log_exception(exc)
            sys.exit(2)

        logdir = stimelogging.get_logfile_dir(recipe.log) or '.'
        log.info(f"recipe logs will be saved under {logdir}")

        filename = os.path.join(logdir, "stimela.recipe.deps")
        stimela.config.CONFIG_DEPS.update(recipe_deps)
        stimela.config.CONFIG_DEPS.save(filename)
        log.info(f"saved recipe dependencies to {filename}")

    # in debug mode, pretty-print the recipe
    if log.isEnabledFor(logging.DEBUG):
        log.debug("---------- prevalidated step follows ----------")
        for line in outer_step.summary(params=params):
            log.debug(line)

    if dry_run:
        log.info("dry run was requested, exiting")
        sys.exit(0)

    start_time = datetime.now()
    def elapsed():
        return str(datetime.now() - start_time).split('.', 1)[0]

    try:
        outputs = outer_step.run()
    except Exception as exc:
        task_stats.save_profiling_stats(outer_step.log, 
            print_depth=profile if profile is not None else stimela.CONFIG.opts.profile.print_depth,
            unroll_loops=stimela.CONFIG.opts.profile.unroll_loops)
        if not isinstance(exc, ScabhaBaseException) or not exc.logged:
            log_exception(StimelaRuntimeError(f"run failed after {elapsed()}", exc, 
                tb=not isinstance(exc, ScabhaBaseException)))
        else:
            log.error("run failed, exiting with error code 1")
        for line in traceback.format_exc().split("\n"):
            log.debug(line)
        outer_step.log.info(f"last log directory was [bold green]{stimelogging.get_logfile_dir(outer_step.log) or '.'}[/bold green]")
        sys.exit(1)

    if outputs and outer_step.log.isEnabledFor(logging.DEBUG):
        outer_step.log.debug(f"run successful after {elapsed()}, outputs follow:")
        for name, value in outputs.items():
            if name in recipe.outputs:
                outer_step.log.debug(f"  {name}: {value}")
    else:
        outer_step.log.info(f"run successful after {elapsed()}")

    task_stats.save_profiling_stats(outer_step.log, 
            print_depth=profile if profile is not None else stimela.CONFIG.opts.profile.print_depth,
            unroll_loops=stimela.CONFIG.opts.profile.unroll_loops)
    
    outer_step.log.info(f"last log directory was [bold green]{stimelogging.get_logfile_dir(outer_step.log) or '.'}[/bold green]")
    return 0
    


from stimela import logger, LOG_FILE, LOG_HOME, utils, CAB_USERNAME


def make_parser(subparsers):
    parser = subparsers.add_parser("clean", help='Convience tools for cleaning up after stimela')

    add = parser.add_argument

    add("-ai", "--all-images", action="store_true",
        help="Remove all images pulled/built by stimela. This include CAB images")

    add("-ab", "--all-base", action="store_true",
        help="Remove all base images")

    add("-ac", "--all-cabs", action="store_true",
        help="Remove all CAB images")

    add("-aC", "--all-containers", action="store_true",
        help="Stop and/or Remove all stimela containers")

    add("-bl", "--build-label", default=CAB_USERNAME,
        help="Label for cab images. All cab images will be named <CAB_LABEL>_<cab name>. The default is $USER")

    parser.set_defaults(func=clean)


def clean(args, conf):
    pass




import glob
import os.path
import fnmatch
import pyparsing
pyparsing.ParserElement.enable_packrat()
from pyparsing import *
from pyparsing import common
from functools import reduce
import operator

from .substitutions import SubstitutionError, SubstitutionContext
from .basetypes import Unresolved, UNSET
from .exceptions import *

import typing
from typing import Dict, List, Any


_parser = None

# see https://stackoverflow.com/questions/43244861/pyparsing-infixnotation-optimization for a cleaner parser with functions

def _not_operator(value):
    return value is UNSET or type(value) is UNSET or not value

_UNARY_OPERATORS = {
    '+':    lambda x: +x,
    '-':    lambda x: -x,
    '~':    lambda x: ~x,
    'not':  _not_operator
}

_BINARY_OPERATORS = {
    '**':   lambda x,y: x ** y,
    '*':   lambda x,y: x * y,
    '/':   lambda x,y: x / y,
    '//':   lambda x,y: x // y,
    '+':   lambda x,y: x + y,
    '-':   lambda x,y: x - y,
    '<<':   lambda x,y: x << y,
    '>>':   lambda x,y: x >> y,
    '&':   lambda x,y: x & y,
    '^':   lambda x,y: x ^ y,
    '|':   lambda x,y: x | y,
    '==':   lambda x,y: x == y,
    '!=':   lambda x,y: x != y,
    '<=':   lambda x,y: x <= y,
    '<':   lambda x,y: x < y,
    '>=':   lambda x,y: x >= y,
    '>':   lambda x,y: x > y,
    # 'is':   lambda x,y: x is y,
    # 'is not':   lambda x,y: x is not y,
    'in':   lambda x,y: x in y,
    'not in':   lambda x,y: x not in y,
    'and':   lambda x,y: x and y,
    'or':   lambda x,y: x or y,
}        


class ResultsHandler(object):
    def evaluate(evaluator):
        pass

    def dump(self):
        pass

    @classmethod
    def pa(cls, s, l, t):
        return cls(*t[0])


class UnaryHandler(ResultsHandler):
    def __init__(self, op, arg):
        self.op, self.arg = op, arg
        assert op in _UNARY_OPERATORS
        self._op = _UNARY_OPERATORS[op]
        # not allows UNSET values
        self.allow_unset = (op == 'not')
    
    def evaluate(self, evaluator):
        arg = evaluator._evaluate_result(self.arg, allow_unset=self.allow_unset)
        if type(arg) is UNSET:
            return arg
        return self._op(arg)

    def dump(self):
        return f"UnaryHandler({self.op}\n   {self.arg})"


class BinaryHandler(ResultsHandler):
    def __init__(self, arg1, op, arg2):
        self.op, self.arg1, self.arg2 = op, arg1, arg2
        assert op in _BINARY_OPERATORS
        self._op = _BINARY_OPERATORS[op]
        self.allow_unset = False
    
    def evaluate(self, evaluator):
        arg1, arg2 = evaluator._evaluate_result(self.arg1, allow_unset=self.allow_unset), \
                     evaluator._evaluate_result(self.arg2, allow_unset=self.allow_unset)
        if type(arg1) is UNSET:
            return arg1
        if type(arg2) is UNSET:
            return arg2
        return self._op(arg1, arg2)

    def dump(self):
        return f"BinaryHandler({self.arg1},\n   {self.op},\n{self.arg2})"

    @staticmethod
    def pa(s, l, t):
        # https://stackoverflow.com/questions/4571441/recursive-expressions-with-pyparsing
        initlen, incr = 3, 2
        t = t[0]
        ret = BinaryHandler(*t[:initlen])
        i = initlen
        while i < len(t):
            ret = BinaryHandler(ret, *t[i:i+incr])
            i += incr
        return ret

class FunctionHandler(ResultsHandler):
    def __init__(self, func, *args):
        self.func, self.args = func, args
        self._func = getattr(self, func)
        if self._func is None:
            raise FormulaError(f"unknown function {func}")

    @staticmethod
    def pa(s, l, t):
        # https://stackoverflow.com/questions/4571441/recursive-expressions-with-pyparsing
        return FunctionHandler(*t[0])

    def evaluate(self, evaluator):
        return self._func(evaluator, self.args)

    def evaluate_generic_callable(self, evaluator, name, callable, args, min_args=None, max_args=None):
        if min_args is not None and len(args) < min_args:
            raise FormulaError(f"{'.'.join(evaluator.location)}: {name}() expects at least {min_args} argument(s)")
        if max_args is not None and len(args) > max_args:
            raise FormulaError(f"{'.'.join(evaluator.location)}: {name}() expects at most {max_args} argument(s)")
        eval_args = [evaluator._evaluate_result(arg) for arg in args]
        # if any argument is UNSET, return it as our result
        unsets = [arg for arg in eval_args if type(arg) is UNSET]
        if unsets:
            return unsets[0]
        return callable(*eval_args)

    def LIST(self, evaluator, args):
        def make_list(*x):
            return list(x)
        return self.evaluate_generic_callable(evaluator, "LIST", make_list, args)

    def MIN(self, evaluator, args):
        return self.evaluate_generic_callable(evaluator, "MIN", min, args, min_args=1)

    def MAX(self, evaluator, args):
        return self.evaluate_generic_callable(evaluator, "MAX", max, args, min_args=1)

    def IF(self, evaluator, args):
        if len(args) < 3 or len(args) > 4:
            raise FormulaError(f"{'.'.join(evaluator.location)}: IF() expects 3 or 4 arguments, got {len(args)}")
        conditional, if_true, if_false = args[:3]
        if_unset = args[3] if len(args) == 4 else UNSET("")

        cond = evaluator._evaluate_result(conditional, allow_unset=if_unset is not None)
        if type(cond) is UNSET:
            if if_unset is None:
                raise SubstitutionError(f"{'.'.join(evaluator.location)}: '{cond.name}' is not defined")
            result = if_unset
        else:
            result = if_true if cond else if_false    
            
        return evaluator._evaluate_result(result)

    def IFSET(self, evaluator, args):
        if len(args) < 1 or len(args) > 3:
            raise FormulaError(f"{'.'.join(evaluator.location)}: IFSET() expects 1 to 3 arguments, got {len(args)}")
        
        lookup, if_set, if_unset = list(args) + [None]*(3 - len(args))

        value = evaluator._evaluate_result(lookup, allow_unset=True)
        if type(value) is UNSET:
            if is_missing(if_unset):
                return UNSET
            else:
                return evaluator._evaluate_result(if_unset)
        elif is_missing(if_set) or if_set == 'SELF':
            return value
        else:
            return evaluator._evaluate_result(if_set)            

    def GLOB(self, evaluator, args):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(evaluator.location)}: GLOB() expects 1 argument, got {len(args)}")
        pattern = evaluator._evaluate_result(args[0])
        if type(pattern) is UNSET:
            return pattern
        return sorted(glob.glob(pattern))

    def EXISTS(self, evaluator, args):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(evaluator.location)}: EXISTS() expects 1 argument, got {len(args)}")
        pattern = evaluator._evaluate_result(args[0])
        if type(pattern) is UNSET:
            return pattern
        return bool(glob.glob(pattern))

    def DIRNAME(self, evaluator, args):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(evaluator.location)}: DIRNAME() expects 1 argument, got {len(args)}")
        path = evaluator._evaluate_result(args[0])
        if type(path) is UNSET:
            return path
        return os.path.dirname(str(path))

    def BASENAME(self, evaluator, args):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(evaluator.location)}: BASENAME() expects 1 argument, got {len(args)}")
        path = evaluator._evaluate_result(args[0])
        if type(path) is UNSET:
            return path
        return os.path.basename(str(path))

    def EXTENSION(self, evaluator, args):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(evaluator.location)}: EXTENSION() expects 1 argument, got {len(args)}")
        path = evaluator._evaluate_result(args[0])
        if type(path) is UNSET:
            return path
        return os.path.splitext(str(path))[1]

    def STRIPEXT(self, evaluator, args):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(evaluator.location)}: STRIPEXT() expects 1 argument, got {len(args)}")
        path = evaluator._evaluate_result(args[0])
        if type(path) is UNSET:
            return path
        return os.path.splitext(str(path))[0]

def construct_parser():
    lparen = Literal("(").suppress()
    rparen = Literal(")").suppress()
    lbrack = Keyword("[").suppress()
    rbrack = Keyword("]").suppress()
    comma = Literal(",").suppress()
    period = Literal(".").suppress()
    string = (QuotedString('"') | QuotedString("'"))("constant")
    UNSET = Keyword("UNSET")("unset")
    SELF = Keyword("SELF")("self_value")
    EMPTY = Keyword("EMPTY")("empty")
    bool_false = (Keyword("False") | Keyword("false"))("bool_false").set_parse_action(lambda:[False])
    bool_true = (Keyword("True") | Keyword("true"))("bool_true").set_parse_action(lambda:[True])

    boolean = (bool_true | bool_false)("constant")
    number = common.number("constant")

    fieldname = Word(alphas + "_", alphanums + "_-@*?")
    nested_field = Group(fieldname + OneOrMore(period + fieldname))("namespace_lookup")
    anyseq = CharsNotIn(",)")("constant")

    # allow expression to be used recursively
    expr = Forward()
    
    # functions
    functions = reduce(operator.or_, map(Keyword, ["IF", "IFSET", "GLOB", "EXISTS", "LIST", 
        "BASENAME", "DIRNAME", "EXTENSION", "STRIPEXT", "MIN", "MAX"]))
    # these functions take one argument, which could also be a sequence
    anyseq_functions = reduce(operator.or_, map(Keyword, ["GLOB", "EXISTS"]))

    atomic_value = (boolean | UNSET | EMPTY | nested_field | string | number)

    function_call_anyseq = Group(anyseq_functions + lparen + anyseq + rparen).setParseAction(FunctionHandler.pa)
    function_call = Group(functions + lparen + 
                    Opt(delimited_list(expr|SELF)) + 
                    rparen).setParseAction(FunctionHandler.pa)
    operators = (
        (Literal("**"), 2, opAssoc.LEFT, BinaryHandler.pa), 
        (Literal("-")|Literal("+")|Literal("~"), 1, opAssoc.RIGHT, UnaryHandler.pa), 
        (Literal("*")|Literal("//")|Literal("/")|Literal("%"), 2, opAssoc.LEFT, BinaryHandler.pa),
        (Literal("+")|Literal("-"), 2, opAssoc.LEFT, BinaryHandler.pa),
        (Literal("<<")|Literal(">>"), 2, opAssoc.LEFT, BinaryHandler.pa),
        (Literal("&"), 2, opAssoc.LEFT, BinaryHandler.pa),
        (Literal("^"), 2, opAssoc.LEFT, BinaryHandler.pa),
        (Literal("|"), 2, opAssoc.LEFT, BinaryHandler.pa),
        (reduce(operator.or_, map(Literal, ("==", "!=", ">", "<", ">=", "<="))), 2, opAssoc.LEFT, BinaryHandler.pa),
        (CaselessKeyword("in")|CaselessKeyword("not in"), 2, opAssoc.LEFT, BinaryHandler.pa),
        (CaselessKeyword("not"), 1, opAssoc.RIGHT, UnaryHandler.pa),
        (CaselessKeyword("and")|CaselessKeyword("or"), 2, opAssoc.LEFT, BinaryHandler.pa),
    )

    infix = infix_notation(atomic_value | function_call | function_call_anyseq | nested_field,
                            operators)("subexpression")

    expr <<= infix

    return expr


if _parser is None:
    _parser = construct_parser()

_parse_cache = {}

def parse_string(text: str, location: List[str] = []):
    """parses a formula, returns ParseResults object. Implements cache.

    Args:
        text (str): formula to parse
        location (List[str], optional): Hierarchical location, used for erro messages
        
    Raises:
        ParserError: on any parse error

    Returns:
        ParseResults: on successful parse
    """
    parse_results = _parse_cache.get(text)
    if parse_results is None:
        try:
            parse_results = _parser.parse_string(text, parse_all=True)
        except Exception as exc:
            parse_results = ParserError(f"{'.'.join(location)}: error parsing formula ({exc})")
        _parse_cache[text] = parse_results

    if isinstance(parse_results, Exception):
        raise parse_results

    return parse_results

def is_missing(result):
    return result is None

class SELF(object):
    pass


class Evaluator(object):
    def __init__(self,  ns: Dict[str, Any], 
                        subst_context: typing.Optional[SubstitutionContext] = None, 
                        allow_unresolved: bool = False,
                        location: List[str] = []):
        self.ns = ns
        self.subst_context = subst_context
        self.location = location
        self.allow_unresolved = allow_unresolved

    def _resolve(self, value, in_formula=True):
        if type(value) is str:
            if in_formula and value == "SELF":
                return SELF
            elif in_formula and value == "UNSET":
                return UNSET
            elif self.subst_context is not None:
                try:
                    value = self.subst_context.evaluate(value, location=self.location)
                except (KeyError, AttributeError) as exc:
                    raise SubstitutionError(f"{value}: invalid key {exc}")
                except Exception as exc:
                    raise SubstitutionError(f"{value}: {exc}")
        return value

    def empty(self, *args):
        return ""

    def unset(self, *args):
        return UNSET

    def self_value(self, *args):
        return SELF

    def constant(self, value):
        return self._resolve(value)

    def subexpression(self, value):
        return self._evaluate_result(value, allow_unset=True)
    
    def namespace_lookup(self, *args):
        if len(args) == 1 and type(args[0]) is ParseResults:
            args = args[0]
            assert args._name == "namespace_lookup"
        value = self.ns
        fields = list(args)
        while fields:
            fld = fields.pop(0)
            # check for wildcards
            if fld not in value and ('*' in fld or '?' in fld):
                names = sorted(fnmatch.filter(value.keys(), fld))
                if names:
                    fld = names[-1]
            # last element allowed to be UNSET, otherwise substitution error
            if fld not in value:
                if fields:
                    raise SubstitutionError(f"{'.'.join(self.location)}: '{fld}' undefined (in '{'.'.join(args)}')")
                else:
                    return UNSET('.'.join(args))
            # this can still throw an error if a nested interpolation is invoked
            try:
                value = value[fld]
            except (KeyError, AttributeError) as exc:
                raise SubstitutionError(f"{'.'.join(self.location)}: '{'.'.join(args)}' unresolved (at '{exc}')")
        return self._resolve(value)

    def _evaluate_result(self, parse_result, allow_unset=False):
        allow_unset = allow_unset or self.allow_unresolved
        # if result is a handler, use evaluate
        if isinstance(parse_result, ResultsHandler):
            value = parse_result.evaluate(self)

        # if result is already a constant, resolve it
        elif type(parse_result) is not ParseResults:
            return self._resolve(parse_result)
        
        # lookup processing method based on name
        else:
            method = parse_result.getName()
            assert method is not None
            if not hasattr(self, method):
                raise ParserError(f"{'.'.join(self.location)}: don't know how to deal with an element of type '{method}'")

            try:
                value = getattr(self, method)(*parse_result)
            except SubstitutionError as exc:
                if allow_unset:
                    return UNSET("", [exc])
                raise

        if type(value) is UNSET and not allow_unset:
            raise UnsetError(f"'{value.value}' undefined")

        return value

    def evaluate(self, value, sublocation: List[str] = []):
        if type(value) is not str:
            return value

        loclen = len(self.location)
        self.location += sublocation

        try:
            if value.startswith("="):
                if value.startswith("=="):
                    return self._resolve(value[1:], in_formula=False)
                else:
                    try:
                        parse_results = parse_string(value[1:])
                    except Exception as exc:
                        raise ParserError(f"{'.'.join(self.location)}: error parsing formula '{value}'", exc)

                    try:
                        return self._evaluate_result(parse_results, allow_unset=True)
                    except Exception as exc:
                        raise FormulaError(f"{'.'.join(self.location)}: evaluation of '{value}' failed", exc, tb=True)
            else:
                return self._resolve(value, in_formula=False)
        finally:
            self.location = self.location[:loclen]
            

    def evaluate_dict(self, params: Dict[str, Any], 
                    corresponding_ns: typing.Optional[Dict[str, Any]] = None, 
                    defaults: Dict[str, Any] = {}, 
                    raise_substitution_errors: bool = True, 
                    verbose: bool =False):
        params = params.copy()
        for name, value in list(params.items()):
            if type(value) is not Unresolved:
                retry = True
                while retry:
                    retry = False
                    if verbose: # or type(value) is UNSET:
                        print(f"{name}: {value} ...")
                    try:
                        new_value = self.evaluate(value, sublocation=[name])
                    except AttributeError as err:
                        if raise_substitution_errors:
                            raise
                        new_value = Unresolved(errors=[err])
                    except SubstitutionError as err:
                        if raise_substitution_errors:
                            raise
                        new_value = Unresolved(errors=[err])
                    if verbose:
                        print(f"{name}: {value} -> {new_value}")
                    # UNSET return means delete or revert to default
                    if new_value is UNSET:
                        # if value is in defaults, try to evaluate that instead
                        if name in defaults and defaults[name] is not UNSET:
                            value = params[name] = defaults[name]
                            if corresponding_ns:
                                corresponding_ns[name] = str(defaults[name])
                            retry = True
                        else: 
                            del params[name]
                            if corresponding_ns and name in corresponding_ns:
                                del corresponding_ns[name]
                    elif new_value is not value and new_value != value:
                        params[name] = new_value
                        if corresponding_ns:
                            corresponding_ns[name] = new_value
        return params

if __name__ == "__main__":
    pass
